from __future__ import annotations

from collections.abc import Iterator

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.utils import datatypes


class MenuMixin(widgets.WidgetMixin):
    def __init__(self, *args, tool_tips_visible=True, **kwargs):
        super().__init__(*args, tool_tips_visible=tool_tips_visible, **kwargs)

    def __iter__(self) -> Iterator[gui.QAction]:
        return iter(self.actions())

    def __len__(self) -> int:
        return len(self.actions())

    def __add__(self, other: gui.QAction):
        self.add(other)
        return self

    def __getitem__(self, item: str) -> gui.QAction:
        for action in self.actions():
            if action.objectName() == item:
                return action
        raise KeyError(f"Action {item} not in menu")

    def add(self, *item: gui.QAction):
        for i in item:
            i.setParent(self)
            self.addAction(i)

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the menu.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def add_separator(self, text: str | None = None) -> widgets.WidgetAction:
        """Adds a separator showing an optional label.

        Args:
            text: Text to show on separator

        Returns:
            Separator action
        """
        separator = widgets.WidgetAction(self)
        if text is None:
            separator.setSeparator(True)
        else:
            label = widgets.Label(
                text, minimum_width=self.minimumWidth(), alignment="center"
            )
            with label.edit_stylesheet() as ss:
                ss.background.setValue("lightgrey")
            separator.setDefaultWidget(label)
            separator.setEnabled(False)
        self.add(separator)
        return separator

    def add_menu(self, menu: widgets.QMenu) -> gui.Action:
        action = gui.Action(text=menu.title(), icon=menu.icon())
        action.setMenu(menu)
        super().addAction(action)
        return action


class Menu(MenuMixin, widgets.QMenu):
    """Menu widget for use in menu bars, context menus, and other popup menus."""


if __name__ == "__main__":
    app = widgets.app()
    menu = Menu("1")
    action = gui.Action(text="test")
    menu.addAction(action)
    menu.show()
    menu.exec(core.Point(200, 200))



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtDesigner, QtWidgets


class PyDesignerCustomWidgetPlugin(
    core.ObjectMixin, QtDesigner.QPyDesignerCustomWidgetPlugin
):
    def name(self):
        return "Hallo"

    def icon(self):
        return "icon"

    def group(self):
        return "group"

    def includeFile(self):
        return "includeFile"

    def isContainer(self):
        return True

    def toolTip(self):
        return "toolTip"

    def whatsThis(self):
        return "whatsThis"

    def initialize(self, interface: QtDesigner.QDesignerFormEditorInterface):
        return super().initialize(interface)

    def get_code_template(self):
        return self.codeTemplate()

    def create_widget(self, parent: QtWidgets.QWidget):
        return self.createWidget(parent)

    def get_dom_xml(self):
        return self.domXml()


if __name__ == "__main__":
    plg = PyDesignerCustomWidgetPlugin()
    new = dir(PyDesignerCustomWidgetPlugin())
    bie = dir(core.Object())
    c = [x for x in new if x not in bie]



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtPrintSupport


class PrintPreviewDialog(widgets.DialogMixin, QtPrintSupport.QPrintPreviewDialog):
    """Dialog for previewing and configuring page layouts for printer output."""

    @classmethod
    def setup_example(cls):
        return None


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    dlg = PrintPreviewDialog()
    dlg.show()



from __future__ import annotations

import enum
import itertools
import logging

from typing import Literal

from prettyqt import constants, core, gui, itemmodels
from prettyqt.utils import bidict


logger = logging.getLogger(__name__)

ColorModeStr = Literal["all", "visible", "seen", "range"]


class ColorMode(enum.IntEnum):
    All = 1
    Visible = 2
    Seen = 3
    Range = 4


COLOR_MODE = bidict(
    all=ColorMode.All,
    visible=ColorMode.Visible,
    seen=ColorMode.Seen,
    range=ColorMode.Range,
)


class SliceColorCategoriesProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model to apply coloring to categories.

    ### Example

    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table[:, :3].proxify.color_categories()
    table.show()
    # or
    indexer = (slice(None), slice(None, 3))
    proxy = itemmodels.SliceColorCategoriesProxyModel(indexer=indexer)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "color_categories"
    ICON = "mdi.palette-outline"

    def __init__(self, *args, **kwargs):
        self._role = constants.DISPLAY_ROLE
        self._color_map = {}
        self.color_generator = itertools.cycle(gui.Palette().iter_colors())
        self._color_none = False
        self._cast_to_str = False
        super().__init__(*args, **kwargs)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not self.indexer_contains(index):
            return super().data(index, role)
        match role:
            case constants.BACKGROUND_ROLE:
                return self.get_color_for_index(index)
            case _:
                return super().data(index, role)

    def get_color_for_index(self, index):
        value = super().data(index, self._role)
        key = str(value) if self._cast_to_str else value
        match value:
            case None if not self._color_none:
                return None
            case _:
                if key in self._color_map:
                    return self._color_map[key]
                new = next(self.color_generator)
                self._color_map[key] = new
                return self._color_map[key]

    def is_none_colored(self) -> bool:
        return self._color_none

    def set_none_colored(self, val: bool):
        with self.change_layout():
            self._color_none = val

    def is_casted_to_str(self) -> bool:
        return self._cast_to_str

    def set_cast_to_str(self, val: bool):
        with self.change_layout():
            self._color_map = {}
            self.color_generator = itertools.cycle(gui.Palette().iter_colors())
            self._cast_to_str = val

    color_none = core.Property(
        bool,
        is_none_colored,
        set_none_colored,
        doc="Whether None-values should also get colored",
    )
    """Color ItemData with value `None`."""

    cast_to_str = core.Property(
        bool,
        is_casted_to_str,
        set_cast_to_str,
        doc="Cast values to string in order to color / group them",
    )
    """Cast all values to a string for deciding whether cells are in same category."""


if __name__ == "__main__":
    import random

    from prettyqt import debugging, widgets

    app = widgets.app()

    val_range = range(0, 50, 10)
    data = dict(
        a=random.sample(val_range, k=5),
        b=random.sample(val_range, k=5),
        c=random.sample(val_range, k=5),
    )
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[:, :].color_categories()
    table.set_size_adjust_policy("content")
    table.setWindowTitle("Color values")
    table.set_icon("mdi.palette")
    # table.show()
    table.adjustSize()
    widget = debugging.ProxyComparerWidget(table.model())

    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt.qt import QtWidgets


class ItemEditorCreatorBase(QtWidgets.QItemEditorCreatorBase):
    """Abstract base class for implementing new item editor creators."""



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict


SwipeDirectionStr = Literal["none", "left", "right", "up", "down"]

SWIPE_DIRECTION: bidict[SwipeDirectionStr, widgets.QSwipeGesture.SwipeDirection] = bidict(
    none=widgets.QSwipeGesture.SwipeDirection.NoDirection,
    left=widgets.QSwipeGesture.SwipeDirection.Left,
    right=widgets.QSwipeGesture.SwipeDirection.Right,
    up=widgets.QSwipeGesture.SwipeDirection.Up,
    down=widgets.QSwipeGesture.SwipeDirection.Down,
)

HorizontalDirectionStr = Literal["none", "left", "right"]
VerticalDirectionStr = Literal["none", "up", "down"]


class SwipeGestureMixin(widgets.GestureMixin):
    def get_horizontal_direction(self) -> HorizontalDirectionStr:
        """Return horizontal direction of the gesture.

        Returns:
            horizontal direction
        """
        return SWIPE_DIRECTION.inverse[self.horizontalDirection()]

    def get_vertical_direction(self) -> VerticalDirectionStr:
        """Return vertical direction of the gesture.

        Returns:
            vertical direction
        """
        return SWIPE_DIRECTION.inverse[self.verticalDirection()]


class SwipeGesture(SwipeGestureMixin, widgets.QSwipeGesture):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    gesture = SwipeGesture()



from __future__ import annotations

from collections.abc import Iterator
from typing import Literal

from prettyqt import constants, widgets
from prettyqt.utils import bidict


mod = widgets.QFormLayout

ItemRoleStr = Literal["left", "right", "both"]

ITEM_ROLE: bidict[ItemRoleStr, mod.ItemRole] = bidict(
    left=mod.ItemRole.LabelRole,
    right=mod.ItemRole.FieldRole,
    both=mod.ItemRole.SpanningRole,
)

RowWrapPolicyStr = Literal["dont_wrap", "wrap_long", "wrap_all"]

ROW_WRAP_POLICY: bidict[RowWrapPolicyStr, mod.RowWrapPolicy] = bidict(
    dont_wrap=mod.RowWrapPolicy.DontWrapRows,
    wrap_long=mod.RowWrapPolicy.WrapLongRows,
    wrap_all=mod.RowWrapPolicy.WrapAllRows,
)

FieldGrowthPolicyStr = Literal[
    "fields_stay_at_size", "expanding_fields_grow", "all_non_fixed_fields_grow"
]

FIELD_GROWTH_POLICY: bidict[FieldGrowthPolicyStr, mod.FieldGrowthPolicy] = bidict(
    fields_stay_at_size=mod.FieldGrowthPolicy.FieldsStayAtSizeHint,
    expanding_fields_grow=mod.FieldGrowthPolicy.ExpandingFieldsGrow,
    all_non_fixed_fields_grow=mod.FieldGrowthPolicy.AllNonFixedFieldsGrow,
)


class FormLayout(widgets.LayoutMixin, widgets.QFormLayout):
    """Manages forms of input widgets and their associated labels."""

    ID = "form"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_size_constraint("maximum")
        self.setVerticalSpacing(8)

    def __setitem__(
        self, index: int | tuple[int, ItemRoleStr], value: str | widgets.QWidget
    ):
        match index:
            case (int() as row, str() as role):
                self.set_widget(value, row, role)
            case int() as row:
                self.set_widget(value, row, "both")
            case _:
                raise TypeError(index)

    def __delitem__(self, index: int):
        self.removeRow(index)

    def __iter__(self) -> Iterator[widgets.QWidget | widgets.QLayout]:
        return iter(self[i] for i in range(self.count()) if self[i] is not None)

    def __len__(self) -> int:
        """Needed for PySide2."""
        return self.rowCount()

    def __add__(self, other: widgets.QWidget | widgets.QLayout | tuple):
        self.add(other)
        return self

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "FieldGrowthPolicy": FIELD_GROWTH_POLICY,
            "formAlignment": constants.ALIGNMENTS,
            "labelAlignment": constants.ALIGNMENTS,
            "rowWrapPolicy": ROW_WRAP_POLICY,
        }
        return maps

    def set_form_alignment(
        self, alignment: constants.AlignmentStr | constants.AlignmentFlag
    ):
        """Set the alignment of the form.

        Args:
            alignment: alignment for the form
        """
        self.setFormAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    def get_form_alignment(self) -> constants.AlignmentStr:
        """Return current form alignment.

        Returns:
            form alignment
        """
        return constants.ALIGNMENTS.inverse[self.formAlignment()]

    def set_label_alignment(
        self, alignment: constants.AlignmentStr | constants.AlignmentFlag
    ):
        """Set the alignment of the label.

        Args:
            alignment: alignment for the label
        """
        self.setFormAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    def get_label_alignment(self) -> constants.AlignmentStr:
        """Return current label alignment.

        Returns:
            label alignment
        """
        return constants.ALIGNMENTS.inverse[self.labelAlignment()]

    def set_widget(
        self,
        widget: str | widgets.QWidget,
        row: int,
        role: ItemRoleStr | mod.ItemRole = "both",
    ):
        widget = widgets.Label(widget) if isinstance(widget, str) else widget
        self.setWidget(row, ITEM_ROLE.get_enum_value(role), widget)

    def get_widget(
        self, row: int, role: ItemRoleStr | mod.ItemRole = "both"
    ) -> widgets.QLayout | widgets.QWidget:
        item = self.itemAt(row, ITEM_ROLE.get_enum_value(role))
        return i if (i := item.widget()) is not None else item.layout()

    def get_item_position(self, index: int) -> tuple[int, ItemRoleStr] | None:
        row, role = self.getItemPosition(index)  # type: ignore
        return None if row == -1 else (row, ITEM_ROLE.inverse[role])

    def add(self, *items):
        for i in items:
            match i:
                case widgets.QWidget() | widgets.QLayout():
                    self.addRow(i)
                case tuple():
                    self.addRow(*i)
                case _:
                    raise TypeError(i)

    def set_row_wrap_policy(self, policy: RowWrapPolicyStr | mod.RowWrapPolicy):
        """Set row wrap policy to use.

        Args:
            policy: row wrap policy to use
        """
        self.setRowWrapPolicy(ROW_WRAP_POLICY.get_enum_value(policy))

    def get_row_wrap_policy(self) -> RowWrapPolicyStr:
        """Return current row wrap policy.

        Returns:
            row wrap policy
        """
        return ROW_WRAP_POLICY.inverse[self.rowWrapPolicy()]

    def set_field_growth_policy(
        self, policy: FieldGrowthPolicyStr | mod.FieldGrowthPolicy
    ):
        """Set field growth policy to use.

        Args:
            policy: field growth policy to use
        """
        self.setFieldGrowthPolicy(FIELD_GROWTH_POLICY.get_enum_value(policy))

    def get_field_growth_policy(self) -> FieldGrowthPolicyStr:
        """Return current field growth policy.

        Returns:
            field growth policy
        """
        return FIELD_GROWTH_POLICY.inverse[self.fieldGrowthPolicy()]


if __name__ == "__main__":
    app = widgets.app()
    layout = FormLayout()
    layout[3] = "hellooo"
    w = widgets.Widget()
    w.set_layout(layout)
    w.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterable

from prettyqt import constants, core, gui
from prettyqt.utils import datatypes


class Shortcut(core.ObjectMixin, gui.QShortcut):
    def __init__(self, *args, **kwargs):
        match args:
            case (str(), *rest):
                args = (gui.KeySequence(args[0]), *rest)
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.key().toString()

    def set_context(
        self, context: constants.ShortcutContextStr | constants.ShortcutContext
    ):
        """Set shortcut context.

        Args:
            context: shortcut context
        """
        self.setContext(constants.SHORTCUT_CONTEXT.get_enum_value(context))

    def get_context(self) -> constants.ShortcutContextStr:
        """Return shortcut context.

        Returns:
            shortcut context
        """
        return constants.SHORTCUT_CONTEXT.inverse[self.context()]

    def set_key(
        self,
        key: datatypes.KeyCombinationType,
    ):
        keysequence = gui.KeySequence(key)
        self.setKey(keysequence)

    def set_keys(self, keys: Iterable[datatypes.KeyCombinationType]):
        keysequences = [gui.KeySequence(key) for key in keys]
        self.setKeys(keysequences)

    def get_key(self) -> gui.KeySequence:
        """Return the shortcut's key sequence.

        Returns:
            Key sequence
        """
        return gui.KeySequence(self.key())

    def get_keys(self) -> list[gui.KeySequence]:
        return [gui.KeySequence(i) for i in self]


if __name__ == "__main__":
    app = gui.app()
    shortcut = Shortcut("enter", None)
    print(shortcut.get_properties())



from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from typing import Literal

from typing_extensions import Self

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict, listdelegators


StandardButtonStr = Literal[
    "cancel",
    "ok",
    "save",
    "open",
    "close",
    "discard",
    "apply",
    "reset",
    "restore_defaults",
    "help",
    "save_all",
    "yes",
    "yes_to_all",
    "no",
    "no_to_all",
    "abort",
    "retry",
    "ignore",
]

STANDARD_BUTTON: bidict[StandardButtonStr, widgets.QDialogButtonBox.StandardButton] = (
    bidict(
        cancel=widgets.QDialogButtonBox.StandardButton.Cancel,
        ok=widgets.QDialogButtonBox.StandardButton.Ok,
        save=widgets.QDialogButtonBox.StandardButton.Save,
        open=widgets.QDialogButtonBox.StandardButton.Open,
        close=widgets.QDialogButtonBox.StandardButton.Close,
        discard=widgets.QDialogButtonBox.StandardButton.Discard,
        apply=widgets.QDialogButtonBox.StandardButton.Apply,
        reset=widgets.QDialogButtonBox.StandardButton.Reset,
        restore_defaults=widgets.QDialogButtonBox.StandardButton.RestoreDefaults,
        help=widgets.QDialogButtonBox.StandardButton.Help,
        save_all=widgets.QDialogButtonBox.StandardButton.SaveAll,
        yes=widgets.QDialogButtonBox.StandardButton.Yes,
        yes_to_all=widgets.QDialogButtonBox.StandardButton.YesToAll,
        no=widgets.QDialogButtonBox.StandardButton.No,
        no_to_all=widgets.QDialogButtonBox.StandardButton.NoToAll,
        abort=widgets.QDialogButtonBox.StandardButton.Abort,
        retry=widgets.QDialogButtonBox.StandardButton.Retry,
        ignore=widgets.QDialogButtonBox.StandardButton.Ignore,
    )
)


RoleStr = Literal[
    "invalid",
    "accept",
    "reject",
    "destructive",
    "action",
    "help",
    "yes",
    "no",
    "apply",
    "reset",
]

ROLES: bidict[RoleStr, widgets.QDialogButtonBox.ButtonRole] = bidict(
    invalid=widgets.QDialogButtonBox.ButtonRole.InvalidRole,
    accept=widgets.QDialogButtonBox.ButtonRole.AcceptRole,
    reject=widgets.QDialogButtonBox.ButtonRole.RejectRole,
    destructive=widgets.QDialogButtonBox.ButtonRole.DestructiveRole,
    action=widgets.QDialogButtonBox.ButtonRole.ActionRole,
    help=widgets.QDialogButtonBox.ButtonRole.HelpRole,
    yes=widgets.QDialogButtonBox.ButtonRole.YesRole,
    no=widgets.QDialogButtonBox.ButtonRole.NoRole,
    apply=widgets.QDialogButtonBox.ButtonRole.ApplyRole,
    reset=widgets.QDialogButtonBox.ButtonRole.ResetRole,
)


class DialogButtonBox(widgets.WidgetMixin, widgets.QDialogButtonBox):
    """Widget presenting buttons in a layout that is appropriate to the widget style."""

    button_clicked = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clicked.connect(self.on_click)

    def __len__(self) -> int:
        return len(self.buttons())

    def __getitem__(self, index: StandardButtonStr) -> widgets.QPushButton:
        return self.button(STANDARD_BUTTON[index])

    def __iter__(self) -> Iterator[widgets.QAbstractButton]:
        return iter(self.buttons())

    def __contains__(self, index: StandardButtonStr):
        return self.button(STANDARD_BUTTON[index]) is not None

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "orientation": constants.ORIENTATION,
            "standardButtons": STANDARD_BUTTON,
        }
        return maps

    @classmethod
    def create(cls, **kwargs) -> Self:
        box = cls()
        for k, v in kwargs.items():
            box.add_default_button(k, callback=v)  # type: ignore
        return box

    def on_click(self, button: core.QObject):
        self.button_clicked.emit(button.objectName())

    def set_horizontal(self):
        self.setOrientation(constants.HORIZONTAL)

    def set_vertical(self):
        self.setOrientation(constants.VERTICAL)

    def set_orientation(
        self, orientation: constants.OrientationStr | constants.Orientation
    ):
        """Set the orientation of the button box.

        Args:
            orientation: orientation for the button box
        """
        self.setOrientation(constants.ORIENTATION.get_enum_value(orientation))

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]

    def add_default_buttons(
        self, buttons: Sequence[StandardButtonStr]
    ) -> listdelegators.ListDelegator[widgets.QPushButton]:
        return [self.add_default_button(btn) for btn in buttons]

    def add_default_button(
        self,
        button: StandardButtonStr | widgets.QDialogButtonBox.StandardButton,
        callback: Callable | None = None,
    ) -> widgets.QPushButton:
        """Add a default button.

        Args:
            button: button to add
            callback: function to call when button gets clicked

        Returns:
            created button
        """
        btn = super().addButton(STANDARD_BUTTON.get_enum_value(button))
        btn.setObjectName(button)
        if callback:
            btn.clicked.connect(callback)
        return btn

    def add_button(
        self,
        button: widgets.QPushButton | str,
        role: RoleStr | widgets.QDialogButtonBox.ButtonRole = "accept",
        callback: Callable | None = None,
    ) -> widgets.QPushButton:
        """Add a button.

        Args:
            button: button to add
            role: role of the button
            callback: function to call when button gets clicked

        Returns:
            created button
        """
        if isinstance(button, str):
            button = widgets.PushButton(button)
        self.addButton(button, ROLES.get_enum_value(role))
        if callback:
            button.clicked.connect(callback)
        return button


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = DialogButtonBox()
    buttons = list(STANDARD_BUTTON.keys())
    widget.add_default_buttons(buttons)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import widgets


class GraphicsItemGroup(widgets.GraphicsItemMixin, widgets.QGraphicsItemGroup):
    """Container that treats a group of items as a single item."""



from __future__ import annotations

from collections.abc import MutableMapping

from prettyqt import core
from prettyqt.qt import QtQml
from prettyqt.utils import datatypes, get_repr


class QmlPropertyMapMixin(core.ObjectMixin):
    pass


class QmlPropertyMap(
    QmlPropertyMapMixin,
    QtQml.QQmlPropertyMap,
    MutableMapping,
    metaclass=datatypes.QABCMeta,
):
    """Allows you to set key-value pairs that can be used in QML bindings."""

    def __repr__(self):
        return get_repr(self, dict(self))

    def __setitem__(self, key: str, value: datatypes.Variant):
        self.insert(key, value)

    def __getitem__(self, key: str) -> datatypes.Variant:
        if not self.contains(key):
            raise KeyError(key)
        return self.value(key)

    def __delitem__(self, key: str):
        self.clear(key)

    def __contains__(self, key):
        return self.contains(key)

    def __bool__(self):
        return not self.isEmpty()

    def __iter__(self):
        return iter(self.keys())


if __name__ == "__main__":
    propmap = QmlPropertyMap()
    propmap["a"] = 2
    propmap.pop("a")
    print(len(propmap))



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.qt import QtPrintSupport
from prettyqt.utils import bidict


PRINT_RANGE = bidict(
    all_pages=QtPrintSupport.QAbstractPrintDialog.PrintRange.AllPages,
    selection=QtPrintSupport.QAbstractPrintDialog.PrintRange.Selection,
    page_range=QtPrintSupport.QAbstractPrintDialog.PrintRange.PageRange,
    current_page=QtPrintSupport.QAbstractPrintDialog.PrintRange.CurrentPage,
)

PrintRangeStr = Literal["all_pages", "selection", "page_range", "current_page"]

PRINT_DIALOG_OPTION = bidict(
    to_file=QtPrintSupport.QAbstractPrintDialog.PrintDialogOption.PrintToFile,
    selection=QtPrintSupport.QAbstractPrintDialog.PrintDialogOption.PrintSelection,
    page_range=QtPrintSupport.QAbstractPrintDialog.PrintDialogOption.PrintPageRange,
    show_page_size=QtPrintSupport.QAbstractPrintDialog.PrintDialogOption.PrintShowPageSize,
    collate_copies=QtPrintSupport.QAbstractPrintDialog.PrintDialogOption.PrintCollateCopies,
    current_page=QtPrintSupport.QAbstractPrintDialog.PrintDialogOption.PrintCurrentPage,
)

PrintDialogOptionStr = Literal[
    "to_file",
    "selection",
    "page_range",
    "show_page_size",
    "collate_copies",
    "current_page",
]


class AbstractPrintDialogMixin(widgets.DialogMixin):
    def get_print_range(self) -> PrintRangeStr:
        return PRINT_RANGE.inverse[self.printRange()]

    def set_print_range(
        self, print_range: PrintRangeStr | QtPrintSupport.QAbstractPrintDialog.PrintRange
    ):
        """Set print range.

        Args:
            print_range: print range
        """
        self.setPrintRange(PRINT_RANGE.get_enum_value(print_range))


class AbstractPrintDialog(AbstractPrintDialogMixin, QtPrintSupport.QAbstractPrintDialog):
    """Base implementation for print dialogs used to configure printers."""



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui, iconprovider, widgets
from prettyqt.utils import datatypes


AreaStr = Literal["top", "bottom"]


class SidebarWidget(widgets.MainWindow):
    SETTINGS_BUTTON_HEIGHT = 28

    def __init__(
        self,
        *args,
        show_settings: bool = False,
        main_layout: widgets.widget.LayoutStr | widgets.QLayout = "vertical",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._button_width = 100
        self._style: constants.ToolButtonStyleStr = "text_below_icon"
        self.button_map: dict[widgets.QWidget, widgets.QToolButton] = {}
        self.icon_map: dict[widgets.QWidget, gui.Icon] = {}
        self.sidebar = widgets.ToolBar(
            context_menu_policy="prevent",
            floatable=True,
            object_name="SidebarWidget",
            window_title="Sidebar",
        )
        self.sidebar.set_style(self._style)
        self.sidebar.set_allowed_areas("all")
        self.settings_menu = widgets.Menu()
        self.sidebar.set_icon_size(int(self._button_width * 0.7))
        if show_settings:
            self.settings_btn = self.sidebar.add_menu_button(
                "", icon="mdi.wrench", menu=self.settings_menu
            )
            self.settings_btn.setFixedSize(
                self._button_width, self.SETTINGS_BUTTON_HEIGHT
            )
            self.settings_btn.set_style("icon")
            self.sidebar.orientationChanged.connect(self._on_orientation_change)
            self.sidebar.add_separator()
        self.spacer_action = self.sidebar.add_spacer()
        self.add_toolbar(self.sidebar, "left")
        self.area = widgets.Widget()
        self.area.set_layout("stacked")
        w = widgets.Widget()
        w.set_layout(main_layout)
        self.main_layout = w.box
        self.main_layout.set_margin(0)
        self.main_layout += self.area
        self.setCentralWidget(w)

    def _on_orientation_change(self, orientation: constants.Orientation):
        if orientation == constants.HORIZONTAL:
            self.settings_btn.setFixedSize(34, 34)
        else:
            self.settings_btn.setFixedSize(
                self._button_width, self.SETTINGS_BUTTON_HEIGHT
            )

    def add_tab(
        self,
        item: widgets.QWidget,
        title: str,
        icon: datatypes.IconType | None = None,
        show: bool = False,
        shortcut: str | None = None,
        area: AreaStr = "top",
    ):
        self.area.box.add(item)
        act = gui.Action(
            text=title,
            icon=icon,
            shortcut=shortcut or "",
            parent=self.sidebar,
            checkable=True,
            triggered=lambda: self.set_tab(item),
        )
        self.addAction(act)
        button = widgets.ToolButton(self.sidebar)
        button.setDefaultAction(act)
        button.setFixedWidth(self._button_width)
        button.set_style(self._style)
        if area == "top":
            self.sidebar.insertWidget(self.spacer_action, button)
        else:
            self.sidebar.addWidget(button)
        if len(self.area.box) == 1:
            button.setChecked(True)
        self.button_map[item] = button
        self.icon_map[item] = iconprovider.get_icon(icon)
        if show:
            self.area.box.setCurrentWidget(item)

    def set_marker(
        self, item: str | int | widgets.Widget, color: datatypes.ColorType = "red"
    ):
        widget = item if isinstance(item, widgets.QWidget) else self._get_widget(item)
        if widget == self._get_current_widget():
            return
        template = self.icon_map[widget]
        px = template.pixmap(100, 100)
        with gui.Painter(px) as painter:
            dot = gui.Pixmap.create_dot(color)
            painter.drawPixmap(0, 0, dot)
        icon = gui.Icon(px)
        self.button_map[widget].setIcon(icon)

    def _get_widget(self, item: str | int | widgets.Widget):
        """Returns widget page specified by name, offset or content."""
        match item:
            case int():
                return self.area.box[item]
            case str():
                w = self.area.find_child(widgets.QWidget, name=item, recursive=False)
                if w not in self.area.box:
                    raise ValueError("Layout does not contain the chosen widget")
                return w
            case _:
                raise TypeError(item)

    def _get_current_widget(self) -> widgets.QWidget:
        """Returns the currently selected widget page."""
        for k, v in self.button_map.items():
            if v.isChecked():
                return k
        raise RuntimeError("no page activated.")

    def set_tab(self, item: str | int | widgets.QWidget):
        widget = item if isinstance(item, widgets.QWidget) else self._get_widget(item)
        current = self.area.box.currentWidget()
        self.button_map[current].setChecked(False)
        self.area.box.setCurrentWidget(widget)
        self.button_map[widget].setChecked(True)

    def add_spacer(self) -> gui.QAction:
        return self.sidebar.add_spacer()

    def add_separator(self, text: str | None = None, area: AreaStr = "top"):
        if area == "top":
            self.sidebar.add_separator(text, before=self.spacer_action)
        else:
            self.sidebar.add_separator(text)

    def add_action(self, *args, area: AreaStr = "top", **kwargs):
        act = args[0] if args else gui.Action(**kwargs)
        self.addAction(act)
        button = widgets.ToolButton(self.sidebar)
        button.setDefaultAction(act)
        button.setFixedWidth(self._button_width)
        button.set_style(self._style)
        if area == "top":
            self.sidebar.insertWidget(self.spacer_action, button)
        else:
            self.sidebar.addWidget(button)
        return act


if __name__ == "__main__":
    from prettyqt.custom_widgets import commandpalette

    app = widgets.app()
    a = commandpalette.CommandPalette()
    ex = SidebarWidget(show_settings=True)
    ex.add_shortcut("Ctrl+P", a.show)
    page_1 = widgets.PlainTextEdit()
    page_2 = widgets.ColorDialog()
    page_3 = widgets.FileDialog()
    ex.add_tab(page_1, "Text", "mdi.timer")
    ex.add_tab(page_2, "Color", "mdi.format-color-fill", area="bottom")
    ex.add_tab(page_3, "Help", "mdi.help-circle-outline")
    ex.set_marker(page_3)
    a.populate_from_widget(ex)
    ex.show()
    app.exec()



"""debugging pachakge."""

import enum
import importlib
import inspect

from prettyqt import core


def get_all_qt_classes(only_qobjects: bool = False):
    import PySide6

    return [
        i[1]
        for namespace in PySide6._find_all_qt_modules()
        for i in inspect.getmembers(importlib.import_module(f"PySide6.{namespace}"))
        if (isinstance(i[1], type) and not only_qobjects)
        or (isinstance(i[1], type) and issubclass(i[1], core.QObject))
    ]


def get_all_qt_enums():
    all_qt_enums = [
        i[1]
        for Klass in get_all_qt_classes()
        for i in inspect.getmembers(Klass)
        if isinstance(i[1], enum.EnumType)
    ]

    all_qt_enums.extend(
        [i[1] for i in inspect.getmembers(core.Qt) if isinstance(i[1], enum.EnumType)]
    )
    return all_qt_enums


def list_all_properties(only_enums: bool = False):
    props = {}
    for klass in get_all_qt_classes(only_qobjects=True):
        metaobj = core.MetaObject(klass.staticMetaObject)
        klass_props = {
            f"{klass.__name__}.{prop.name()}": prop
            for prop in metaobj.get_properties(include_super=False)
        }
        if only_enums:
            klass_props = {
                k: v for k, v in klass_props.items() if v.get_enumerator_type()
            }
        props |= klass_props
    return props


if __name__ == "__main__":
    import pprint

    pprint.pprint(list_all_properties(only_enums=True))



from __future__ import annotations

from typing import Literal

from prettyqt import core, network
from prettyqt.utils import bidict, get_repr


mod = network.QAbstractSocket

BindModeStr = Literal[
    "share_address",
    "dont_share_address",
    "reuse_address_hint",
    "default_for_platform",
]

BIND_MODE: bidict[BindModeStr, mod.BindFlag] = bidict(
    share_address=mod.BindFlag.ShareAddress,
    dont_share_address=mod.BindFlag.DontShareAddress,
    reuse_address_hint=mod.BindFlag.ReuseAddressHint,
    default_for_platform=mod.BindFlag.DefaultForPlatform,
)

NetworkLayerProtocolStr = Literal["ipv4", "ipv6", "any_ip", "unknown"]

NETWORK_LAYER_PROTOCOL: bidict[NetworkLayerProtocolStr, mod.NetworkLayerProtocol] = (
    bidict(
        ipv4=mod.NetworkLayerProtocol.IPv4Protocol,
        ipv6=mod.NetworkLayerProtocol.IPv6Protocol,
        any_ip=mod.NetworkLayerProtocol.AnyIPProtocol,
        unknown=mod.NetworkLayerProtocol.UnknownNetworkLayerProtocol,
    )
)

PauseModeStr = Literal["never", "on_ssl_errors"]

PAUSE_MODES: bidict[PauseModeStr, mod.PauseMode] = bidict(
    never=mod.PauseMode.PauseNever,
    on_ssl_errors=mod.PauseMode.PauseOnSslErrors,
)

mod = mod

SocketErrorStr = Literal[
    "connection_refused",
    "remote_host_closed",
    "host_not_found",
    "socket_access",
    "socket_resource",
    "socket_timeout",
    "diagram_too_large",
    "network",
    "address_in_use",
    "socket_address_not_available",
    "unsupported_socket_operation",
    "proxy_authentication_required",
    "ssl_handshake_failed",
    "unfinished_socket_operation",
    "proxy_connection_refused",
    "proxy_connection_closed",
    "proxy_connection_timeout",
    "proxy_not_found",
    "proxy_protocol",
    "operation",
    "ssl_internal",
    "ssl_invalid_user_data",
    "temporary",
    "unknown_socket",
]

SOCKET_ERROR: bidict[SocketErrorStr, mod.SocketError] = bidict(
    connection_refused=mod.SocketError.ConnectionRefusedError,
    remote_host_closed=mod.SocketError.RemoteHostClosedError,
    host_not_found=mod.SocketError.HostNotFoundError,
    socket_access=mod.SocketError.SocketAccessError,
    socket_resource=mod.SocketError.SocketResourceError,
    socket_timeout=mod.SocketError.SocketTimeoutError,
    diagram_too_large=mod.SocketError.DatagramTooLargeError,
    network=mod.SocketError.NetworkError,
    address_in_use=mod.SocketError.AddressInUseError,
    socket_address_not_available=mod.SocketError.SocketAddressNotAvailableError,
    unsupported_socket_operation=mod.SocketError.UnsupportedSocketOperationError,
    proxy_authentication_required=mod.SocketError.ProxyAuthenticationRequiredError,
    ssl_handshake_failed=mod.SocketError.SslHandshakeFailedError,
    unfinished_socket_operation=mod.SocketError.UnfinishedSocketOperationError,
    proxy_connection_refused=mod.SocketError.ProxyConnectionRefusedError,
    proxy_connection_closed=mod.SocketError.ProxyConnectionClosedError,
    proxy_connection_timeout=mod.SocketError.ProxyConnectionTimeoutError,
    proxy_not_found=mod.SocketError.ProxyNotFoundError,
    proxy_protocol=mod.SocketError.ProxyProtocolError,
    operation=mod.SocketError.OperationError,
    ssl_internal=mod.SocketError.SslInternalError,
    ssl_invalid_user_data=mod.SocketError.SslInvalidUserDataError,
    temporary=mod.SocketError.TemporaryError,
    unknown_socket=mod.SocketError.UnknownSocketError,
)

SocketOptionStr = Literal[
    "low_delay",
    "keep_alive",
    "multicast_ttl",
    "multicast_loopback",
    "type_of_service",
    "send_buffer_size_socket",
    "receive_buffer_size",
    "path_mtu_socket",
]

SOCKET_OPTION: bidict[SocketOptionStr, mod.SocketOption] = bidict(
    low_delay=mod.SocketOption.LowDelayOption,
    keep_alive=mod.SocketOption.KeepAliveOption,
    multicast_ttl=mod.SocketOption.MulticastTtlOption,
    multicast_loopback=mod.SocketOption.MulticastLoopbackOption,
    type_of_service=mod.SocketOption.TypeOfServiceOption,
    send_buffer_size_socket=mod.SocketOption.SendBufferSizeSocketOption,
    receive_buffer_size=mod.SocketOption.ReceiveBufferSizeSocketOption,
    path_mtu_socket=mod.SocketOption.PathMtuSocketOption,
)

SocketStateStr = Literal[
    "unconnected",
    "host_lookup",
    "connecting",
    "connected",
    "bound",
    "closing",
    "listening",
]

SOCKET_STATE: bidict[SocketStateStr, mod.SocketState] = bidict(
    unconnected=mod.SocketState.UnconnectedState,
    host_lookup=mod.SocketState.HostLookupState,
    connecting=mod.SocketState.ConnectingState,
    connected=mod.SocketState.ConnectedState,
    bound=mod.SocketState.BoundState,
    closing=mod.SocketState.ClosingState,
    listening=mod.SocketState.ListeningState,
)

SocketTypeStr = Literal["tcp", "udp", "sctp", "unknown"]

SOCKET_TYPE: bidict[SocketTypeStr, mod.SocketType] = bidict(
    tcp=mod.SocketType.TcpSocket,
    udp=mod.SocketType.UdpSocket,
    sctp=mod.SocketType.SctpSocket,
    unknown=mod.SocketType.UnknownSocketType,
)

TypeOfServiceStr = Literal[
    "network_control",
    "internetwork_control",
    "critic_ecp",
    "flash_override",
    "flash",
    "immediate",
    "priority",
    "routine",
]

TYPE_OF_SERVICE: bidict[TypeOfServiceStr, int] = bidict(
    network_control=224,
    internetwork_control=192,
    critic_ecp=160,
    flash_override=128,
    flash=96,
    immediate=64,
    priority=32,
    routine=0,
)


class AbstractSocketMixin(core.IODeviceMixin):
    def __repr__(self):
        return get_repr(self)

    def bind_to(
        self,
        address: str | network.QHostAddress,
        port: int = 0,
        bind_mode: (
            network.QAbstractSocket.BindFlag | BindModeStr
        ) = "default_for_platform",
    ) -> bool:
        if isinstance(address, str):
            address = network.QHostAddress(address)
        mode = BIND_MODE.get_enum_value(bind_mode)
        return self.bind(address, port, mode)

    def connect_to_host(
        self,
        hostname: str,
        port: int,
        open_mode: core.QIODevice.OpenModeFlag | core.iodevice.OpenModeStr = "read_write",
        protocol: (
            network.QAbstractSocket.NetworkLayerProtocol | NetworkLayerProtocolStr
        ) = "any_ip",
    ):
        mode = core.iodevice.OPEN_MODES.get_enum_value(open_mode)
        prot = NETWORK_LAYER_PROTOCOL.get_enum_value(protocol)
        self.connectToHost(hostname, port, mode, prot)

    def get_error(self) -> SocketErrorStr:
        return SOCKET_ERROR.inverse[self.error()]

    def set_pause_mode(self, mode: PauseModeStr | mod.PauseMode):
        """Set pause mode.

        Args:
            mode: pause mode
        """
        self.setPauseMode(PAUSE_MODES.get_enum_value(mode))

    def get_pause_mode(self) -> PauseModeStr:
        return PAUSE_MODES.inverse[self.pauseMode()]

    def get_proxy(self) -> network.NetworkProxy:
        return network.NetworkProxy(self.proxy())

    # def set_socket_option(self, name: str, value):
    #     self.setSocketOption(SOCKET_OPTION.get_enum_value(name), value)

    # def get_socket_option(self, name: str):
    #     return self.socketOption(SOCKET_OPTION.get_enum_value(name))

    # def set_type_of_service(self, typ: str):
    #     self.set_socket_option("type_of_service", TYPE_OF_SERVICE.get_enum_value(typ))

    # def get_type_of_service(self):
    #     opt = self.get_socket_option("type_of_service")
    #     return TYPE_OF_SERVICE.inverse[opt]

    def get_socket_type(self) -> SocketTypeStr:
        return SOCKET_TYPE.inverse[self.socketType()]

    def get_state(self) -> SocketStateStr:
        return SOCKET_STATE.inverse[self.state()]

    def get_local_address(self) -> network.HostAddress:
        return network.HostAddress(self.localAddress())


class AbstractSocket(AbstractSocketMixin, network.QAbstractSocket):
    """The base functionality common to all socket types."""



from __future__ import annotations

from collections.abc import Sequence
import logging

from prettyqt import constants, core, gui, itemmodels


logger = logging.getLogger(__name__)


class NameColumn(itemmodels.ColumnItem):
    name = "Name"
    doc = "Action name"
    editable = True

    def get_data(
        self,
        item: gui.QAction,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                return item.text()
            case constants.DECORATION_ROLE:
                return item.icon()

    def set_data(
        self,
        item: gui.QAction,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ):
        match role:
            case constants.EDIT_ROLE:
                item.setText(value)
                return True


class ToolTipColumn(itemmodels.ColumnItem):
    name = "ToolTip"
    doc = "ToolTip"
    editable = True

    def get_data(
        self,
        item: gui.QAction,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                return item.toolTip()

    def set_data(
        self,
        item: gui.QAction,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ):
        match role:
            case constants.EDIT_ROLE:
                item.setToolTip(value)
                return True


class ShortcutColumn(itemmodels.ColumnItem):
    name = "Shortcut"
    editable = True

    def get_data(
        self,
        item: gui.QAction,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE:
                return item.shortcut().toString()
            case constants.EDIT_ROLE:
                return item.shortcut()

    def set_data(
        self,
        item: gui.QAction,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ):
        match role:
            case constants.EDIT_ROLE:
                item.setShortcut(value)
                return True


class PriorityColumn(itemmodels.ColumnItem):
    name = "Priority"
    editable = True

    def get_data(
        self,
        item: gui.QAction,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE:
                return gui.action.PRIORITIES.inverse[item.priority()]
            case constants.EDIT_ROLE:
                return item.priority()

    def set_data(
        self,
        item: gui.QAction,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ):
        match role:
            case constants.EDIT_ROLE:
                item.setPriority(value)
                return True


class CheckStateColumn(itemmodels.ColumnItem):
    name = "CheckState"

    def get_data(
        self,
        item: gui.QAction,
        role: constants.ItemDataRole = constants.CHECKSTATE_ROLE,
    ):
        match role:
            case constants.CHECKSTATE_ROLE:
                return (
                    self.to_checkstate(item.isChecked()) if item.isCheckable() else None
                )
        return None

    def set_data(
        self,
        item: gui.QAction,
        value,
        role: constants.ItemDataRole = constants.CHECKSTATE_ROLE,
    ):
        match role:
            case constants.CHECKSTATE_ROLE:
                item.setChecked(not item.isChecked())
                return True
        return False

    def get_flags(self, item):
        DEFAULT = constants.IS_SELECTABLE | constants.IS_ENABLED
        return DEFAULT | constants.IS_CHECKABLE if item.isCheckable() else DEFAULT


class UsageCountColumn(itemmodels.ColumnItem):
    name = "Usage count"

    def get_data(
        self, item: gui.QAction, role: constants.ItemDataRole = constants.DISPLAY_ROLE
    ):
        match role:
            case constants.DISPLAY_ROLE if hasattr(item, "usage_count"):
                return item.usage_count


class ActionsModel(itemmodels.ColumnTableModel):
    """Table model to display a list of QActions.

    All properties of the Action can be edited.
    """

    SUPPORTS = Sequence[gui.QAction]
    COLUMNS = [
        NameColumn,
        ToolTipColumn,
        ShortcutColumn,
        CheckStateColumn,
        PriorityColumn,
        UsageCountColumn,
    ]

    def __init__(self, actions: list[gui.QAction], parent: core.QObject | None = None):
        super().__init__(actions, self.COLUMNS, parent=parent)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (gui.QAction(), *_):
                return True
            case _:
                return False


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    actions = [
        gui.Action(
            text="Some nice Qt Action",
            shortcut="Ctrl+A",
            tool_tip="wonderful tooltip.",
            icon="mdi.folder",
            triggered=lambda: print("test"),
        ),
        gui.Action(
            text="Another Qt action",
            shortcut="Ctrl+B",
            tool_tip="wonderful tooltip.",
            icon="mdi.folder-outline",
        ),
        gui.Action(
            text="...and another one",
            shortcut="Ctrl+Alt+A",
            tool_tip="wonderful tooltip.",
            icon="mdi.information",
            checkable=True,
        ),
        gui.Action(
            text="....even more!",
            shortcut="Ctrl+Alt+C",
            tool_tip="wonderful tooltip.",
            icon="mdi.download",
            checkable=True,
            checked=True,
        ),
    ]
    # actions[0].trigger()
    # actions[0].trigger()
    # actions[0].trigger()
    # actions[1].trigger()
    model = ActionsModel(actions, parent=view)
    view.setModel(model)
    view.resize(640, 480)
    view.set_delegate("editor")
    view.set_selection_behavior("rows")
    view.adapt_sizes()
    view.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr


PerformanceHintStr = Literal["moderate", "aggressive"]

PERFORMANCE_HINT: bidict[PerformanceHintStr, QtGui.QStaticText.PerformanceHint] = bidict(
    moderate=QtGui.QStaticText.PerformanceHint.ModerateCaching,
    aggressive=QtGui.QStaticText.PerformanceHint.AggressiveCaching,
)


class StaticText(QtGui.QStaticText):
    def __repr__(self):
        return get_repr(self, self.text())

    def __str__(self):
        return self.text()

    def get_size(self) -> core.SizeF:
        return core.SizeF(self.size())

    def set_text_format(
        self, text_format: constants.TextFormatStr | constants.TextFormat
    ):
        """Set the text format.

        Allowed values are "rich", "plain", "auto", "markdown"

        Args:
            text_format: text format to use
        """
        self.setTextFormat(constants.TEXT_FORMAT.get_enum_value(text_format))

    def get_text_format(self) -> constants.TextFormatStr:
        """Return current text format.

        Possible values: "rich", "plain", "auto", "markdown"

        Returns:
            text format
        """
        return constants.TEXT_FORMAT.inverse[self.textFormat()]

    def set_performance_hint(
        self, hint: PerformanceHintStr | QtGui.QStaticText.PerformanceHint
    ):
        """Set the performance hint.

        Args:
            hint: performance hint to use
        """
        self.setPerformanceHint(PERFORMANCE_HINT.get_enum_value(hint))

    def get_performance_hint(self) -> PerformanceHintStr:
        """Return current performance hint.

        Returns:
            performance hint
        """
        return PERFORMANCE_HINT.inverse[self.performanceHint()]


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    text = StaticText()
    text.get_size()



from __future__ import annotations

from prettyqt import core


class TransposeProxyModel(core.AbstractProxyModelMixin, core.QTransposeProxyModel):
    """This proxy transposes the source model."""

    ID = "transpose"



from __future__ import annotations

from typing import Literal

from prettyqt import network
from prettyqt.utils import bidict, get_repr


mod = network.QHostAddress

CONVERSION_MODE = bidict(
    strict=mod.ConversionModeFlag.StrictConversion,
    v4_mapped_to_ipv4=mod.ConversionModeFlag.ConvertV4MappedToIPv4,
    v4_compat_to_ipv4=mod.ConversionModeFlag.ConvertV4CompatToIPv4,
    localhost=mod.ConversionModeFlag.ConvertLocalHost,
    unspecified_address=mod.ConversionModeFlag.ConvertUnspecifiedAddress,
    tolerant=mod.ConversionModeFlag.TolerantConversion,
)

SPECIAL_ADDRESS = bidict(
    null=mod.SpecialAddress.Null,
    localhost=mod.SpecialAddress.LocalHost,
    localhost_ipv6=mod.SpecialAddress.LocalHostIPv6,
    broadcast=mod.SpecialAddress.Broadcast,
    any_ipv4=mod.SpecialAddress.AnyIPv4,
    any_ipv6=mod.SpecialAddress.AnyIPv6,
    any=mod.SpecialAddress.Any,
)

SpecialAddressStr = Literal[
    "null", "localhost", "localhost_ipv6", "broadcast", "any_ipv4", "any_ipv6", "any"
]

NetworkLayerProtocolStr = Literal["ipv4", "ipv6", "any_ip", "unknown"]


class HostAddress(network.QHostAddress):
    """IP address."""

    def __repr__(self):
        return get_repr(self, self.toString())

    def __str__(self):
        return self.toString()

    def __bool__(self):
        return not self.isNull()

    def get_protocol(self) -> NetworkLayerProtocolStr:
        return network.abstractsocket.NETWORK_LAYER_PROTOCOL.inverse[self.protocol()]

    def set_address(self, address: int | str):
        if address in SPECIAL_ADDRESS:
            address = SPECIAL_ADDRESS[address]
        ret = self.setAddress(address)
        if ret is False:
            raise ValueError("invalid address")



from __future__ import annotations

from prettyqt import core, widgets


MAX_VAL = 2147483647


class SpinBox(widgets.AbstractSpinBoxMixin, widgets.QSpinBox):
    value_changed = core.Signal(int)

    def __init__(self, *args, maximum: int = MAX_VAL, **kwargs):
        super().__init__(*args, maximum=maximum, **kwargs)
        self.valueChanged.connect(self.value_changed)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, int)

    def set_range(self, start: int | None, end: int | None):
        self.setMinimum(start)
        self.setMaximum(end)

    setRange = set_range

    def set_minimum(self, value: int | None):
        if value is None:
            value = -MAX_VAL
        super().setMinimum(value)

    setMinimum = set_minimum

    def set_maximum(self, value: int | None):
        if value is None:
            value = MAX_VAL
        super().setMaximum(value)

    setMaximum = set_maximum

    def set_step_size(self, step_size):
        self.setSingleStep(step_size)


if __name__ == "__main__":
    app = widgets.app()
    widget = SpinBox()
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, core, eventfilters


if TYPE_CHECKING:
    from prettyqt import widgets


class AutoSizeColumnsEventFilter(eventfilters.BaseEventFilter):
    ID = "autosize_columns"

    def __init__(
        self,
        parent: widgets.TableView | widgets.TreeView,
        orientation: constants.Orientation
        | constants.OrientationStr = constants.VERTICAL,
    ):
        super().__init__(parent)
        self._widget = parent
        self.orientation = constants.ORIENTATION.get_enum_value(orientation)
        parent.model_changed.connect(self._on_model_change)
        self._autosized_sections = set()
        self.last_span: tuple[int, int] | None = None

        if self.orientation == constants.VERTICAL:
            parent.h_scrollbar.valueChanged.connect(self._on_scroll)
        else:
            parent.v_scrollbar.valueChanged.connect(self._on_scroll)
        # if sel_model := parent.selectionModel():
        #     sel_model.currentColumnChanged.connect(self._resize_current_col_to_content)

    def _on_model_change(self):
        self._autosized_sections = set()
        # sel_model = self._widget.selectionModel()
        # sel_model.currentColumnChanged.connect(self._resize_current_col_to_content)

    def eventFilter(self, obj, event: core.Event) -> bool:
        match event.type():
            case core.Event.Type.Resize:
                self._on_scroll()
                return False
        return super().eventFilter(obj, event)

    # def _resize_current_col_to_content(self, new_index, old_index):
    #     if new_index.column() not in self._autosized_sections:
    #         # ensure the requested column is fully into view after resizing
    #         self._widget.resize_visible_columns_to_contents()
    #         self._widget.scrollTo(new_index)

    def _on_scroll(self):
        if self.orientation == constants.VERTICAL:
            colcount = self._widget.model().columnCount()
            span = self._widget.get_visible_section_span("horizontal")
            if span == self.last_span:
                return
            self.last_span = span
            col, end = span
            width = self._widget.viewport().width()
            while col <= end:
                if col not in self._autosized_sections:
                    self._autosized_sections.add(col)
                    self._widget.resizeColumnToContents(col)
                col += 1
                end = self._widget.columnAt(width)
                end = colcount if end == -1 else end
        else:
            rowcount = self._widget.model().rowCount()
            span = self._widget.get_visible_section_span("vertical")
            if span == self.last_span:
                return
            self.last_span = span
            row, end = span
            height = self._widget.viewport().height()
            while row <= end:
                if row not in self._autosized_sections:
                    self._autosized_sections.add(row)
                    self._widget.resizeRowToContents(row)
                row += 1
                end = self._widget.rowAt(height)
                end = rowcount if end == -1 else end


if __name__ == "__main__":
    from prettyqt import debugging, widgets  # noqa: F811

    app = widgets.app()
    widget = debugging.example_table()
    widget.set_delegate("editor")
    test = AutoSizeColumnsEventFilter(widget)
    widget.installEventFilter(test)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets


PM = widgets.QStyle.PixelMetric
SH = widgets.QStyle.StyleHint


class AutoResizeTextEditMixin:
    def __init__(self, parent=None):
        super().__init__(parent)
        self.set_line_wrap_mode("none")
        doc_layout = self.document().documentLayout()
        doc_layout.documentSizeChanged.connect(self.update_vertical_size)
        self.update_vertical_size()

    def update_vertical_size(self):
        # can vertical or horizontal scrollbars appear?
        vcan = self.get_vertical_scrollbar_policy() == "as_needed"
        hcan = self.get_horizontal_scrollbar_policy() == "as_needed"

        # width a scrollbar takes off the viewport size
        frame_width = 0
        style = self.style()
        if style.styleHint(SH.SH_ScrollView_FrameOnlyAroundContents, None, self):
            frame_width = style.pixelMetric(PM.PM_DefaultFrameWidth) * 2
        extent = style.pixelMetric(PM.PM_ScrollBarExtent, None, self) + frame_width
        doc = self.document()
        margin = frame_width + doc.documentMargin()
        size = self.get_document_size() + core.QSize(int(margin), int(margin))
        width = size.width()
        height = size.height()
        max_width = self.width()
        max_height = self.maximumHeight()
        # will scrollbars appear?
        hwill, vwill = False, False
        if hcan and width > max_width:
            hwill = True
        if vcan and height > max_height:
            vwill = True
        if vcan and hwill and height + extent > max_height:
            vwill = True
        if hcan and vwill and width + extent > max_width:
            hwill = True
        if hwill:
            height += extent
        self.resize(self.width(), min(max_height, height))

    def get_document_size(self):
        """Implemented differently for QTextEdit and QPlainTextEdit."""
        raise NotImplementedError

    def setLineWrapMode(self, mode):
        """Reimplemented to avoid WidgetWidth wrap mode, which causes resize loops."""
        if mode == self.LineWrapMode.WidgetWidth:
            raise ValueError("cannot use WidgetWidth wrap mode")
        else:
            super().setLineWrapMode(mode)


class AutoResizeTextEdit(AutoResizeTextEditMixin, widgets.TextEdit):
    """TextEdit which adjusts its height to the contained text."""

    def get_document_size(self) -> core.QSize:
        return self.document().documentLayout().documentSize().toSize()


class AutoResizePlainTextEdit(AutoResizeTextEditMixin, widgets.PlainTextEdit):
    """PlainTextEdit which adjusts its height to the contained text."""

    def get_document_size(self) -> core.QSize:
        doc = self.document()
        layout = doc.documentLayout()
        size = layout.documentSize().toSize()
        block = doc.firstBlock()
        line_height = layout.blockBoundingRect(block).height() / block.lineCount()
        height = int(size.height() * line_height + 2 * doc.documentMargin())
        return core.QSize(size.width(), height)


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.Widget()
    layout = widget.set_layout("grid")
    textedit = AutoResizePlainTextEdit()
    layout[0, 1] = textedit
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class SmoothScrollBar(widgets.ScrollBar):
    scroll_ended = core.Signal()

    def __init__(
        self,
        orientation: constants.Orientation | constants.OrientationStr = "horizontal",
        parent: widgets.QAbstractScrollArea | None = None,
        animation_duration: int = 500,
        easing: core.easingcurve.TypeStr | core.QEasingCurve.Type = "out_cubic",
        trigger: bool = False,
    ):
        super().__init__(orientation, parent)
        self._value = self.value()
        self.widget = parent
        self._animation = core.PropertyAnimation()
        self._animation.apply_to(self.value)
        self._animation.set_easing(easing)
        self._animation.setDuration(animation_duration)
        self._animation.finished.connect(self.scroll_ended)
        self.widget.viewport().installEventFilter(self)
        if trigger:
            self.widget.h_scrollbar.valueChanged.connect(gui.Cursor.fake_mouse_move)

    def mouseMoveEvent(self, e):
        self._animation.stop()
        self._value = self.value()
        super().mouseMoveEvent(e)

    def mousePressEvent(self, e):
        self._animation.stop()
        self._value = self.value()
        super().mousePressEvent(e)

    def mouseReleaseEvent(self, e):
        self._animation.stop()
        self._value = self.value()
        super().mouseReleaseEvent(e)

    def setValue(self, value: int):
        if value == self.value():
            return
        self._animation.stop()
        self.scroll_ended.emit()
        self._animation.set_range(self.value(), value)
        self._animation.start()

    def scroll_by_value(self, value: int):
        """Scroll by given distance."""
        self._value += value
        self._value = min(max(self.minimum(), self._value), self.maximum())
        self.setValue(self._value)

    def scroll_to(self, value: int):
        """Scroll to given position."""
        self._value = value
        self._value = min(max(self.minimum(), self._value), self.maximum())
        self.setValue(self._value)

    def reset_value(self, value):
        self._value = value

    def eventFilter(self, source, event) -> bool:
        if event.type() == core.Event.Type.Wheel and source == self.widget.viewport():
            self.widget.v_scrollbar.scroll_by_value(-event.angleDelta().y())
            return True
        return False

    def set_animation_duration(self, duration: int):
        self._animation.setDuration(duration)

    def get_animation_duration(self) -> int:
        return self._animation.duration()

    def set_animation_easing(
        self,
        easing: core.easingcurve.TypeStr | core.QEasingCurve.Type = "out_cubic",
    ):
        self._animation.set_easing(easing)

    def get_animation_easing(self) -> core.easingcurve.TypeStr:
        return self._animation.get_easing()

    def animationEasing(self) -> core.QEasingCurve.Type:
        return self._animation.type()

    animation_duration = core.Property(
        int,
        set_animation_duration,
        get_animation_duration,
        doc="Duration for the animation",
    )
    animation_easing = core.Property(
        core.EasingCurve.Type,
        set_animation_easing,
        get_animation_easing,
        doc="Animation easing type",
    )


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("fusion")
    widget = widgets.PlainTextEdit("\n".join(f"abc{i}" for i in range(1000)))
    scrollbar = SmoothScrollBar("vertical", parent=widget)
    widget.setVerticalScrollBar(scrollbar)
    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from collections.abc import Callable
import re

from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict, classhelpers, colors, datatypes, get_repr


EchoModeStr = Literal["normal", "no_echo", "password", "echo_on_edit"]

ECHO_MODE: bidict[EchoModeStr, widgets.QLineEdit.EchoMode] = bidict(
    normal=widgets.QLineEdit.EchoMode.Normal,
    no_echo=widgets.QLineEdit.EchoMode.NoEcho,
    password=widgets.QLineEdit.EchoMode.Password,
    echo_on_edit=widgets.QLineEdit.EchoMode.PasswordEchoOnEdit,
)

ActionPositionStr = Literal["leading", "trailing"]

ACTION_POSITION: bidict[ActionPositionStr, widgets.QLineEdit.ActionPosition] = bidict(
    leading=widgets.QLineEdit.ActionPosition.LeadingPosition,
    trailing=widgets.QLineEdit.ActionPosition.TrailingPosition,
)

ValidatorStr = Literal[
    "double",
    "color",
    "integer",
    "whitelist",
    "blacklist",
    "integer_classic",
    "path",
    "scientific_integer",
    "scientific_float",
    "python_code",
    "not_zero",
    "json",
    "regex",
    "regular_expression",
    "regex_pattern",
    "int_list",
    "float_list",
    "qss",
    "hex",
    "alphanumeric",
    "text_length",
    "website",
    "email",
]

WEB_REGEX = (
    r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}"
    r"\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)"
)

MAIL_REGEX = r"[a-zA-Z0-9_\.\+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-\.]+"


def get_validator(
    validator: ValidatorStr | datatypes.PatternType,
    **kwargs,
) -> gui.QValidator:
    from prettyqt import validators

    match validator:
        case "email":
            return gui.RegularExpressionValidator(MAIL_REGEX)
        case "website":
            return gui.RegularExpressionValidator(WEB_REGEX)
        case str():
            ValidatorClass = classhelpers.get_class_for_id(gui.ValidatorMixin, validator)
            validator = ValidatorClass(**kwargs)
            return validator
        case core.QRegularExpression():
            return gui.RegularExpressionValidator(validator)
        case re.Pattern():
            return gui.RegularExpressionValidator(core.RegularExpression(validator))
        case Callable():
            return validators.FunctionValidator(validator)
        case _:
            raise ValueError(validator)


class LineEdit(widgets.WidgetMixin, widgets.QLineEdit):
    """One-line text editor."""

    value_changed = core.Signal(str)
    tab_pressed = core.Signal()

    def __init__(self, *args, **kwargs):
        self._validation_color = gui.Color("orange")
        super().__init__(*args, **kwargs)
        self.textChanged.connect(self._on_value_change)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, str)

    def _on_value_change(self):
        val = self.get_value()
        self._update_background()
        self.value_changed.emit(val)

    def __repr__(self):
        return get_repr(self, self.text())

    def __add__(self, other: str):
        self.append_text(other)
        return self

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "echoMode": ECHO_MODE,
            "cursorMoveStyle": constants.CURSOR_MOVE_STYLE,
            "alignment": constants.ALIGNMENTS,
        }
        return maps

    def font(self) -> gui.Font:
        return gui.Font(super().font())

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.key() == constants.Key.Key_Tab:
            self.tab_pressed.emit()

    def append_text(self, text: str):
        self.set_text(self.text() + text)

    def set_text(self, text: str):
        self.setText(text)

    def set_completer(self, completer: widgets.QCompleter | Literal["files"]):
        match completer:
            case widgets.QCompleter():
                self.setCompleter(completer)
            case "files":
                model = widgets.FileSystemModel()
                model.set_root_path("")
                completer = widgets.Completer(self)
                completer.setModel(model)
                self.setCompleter(completer)

    def set_read_only(self, value: bool = True):
        """Set text to read-only.

        Args:
            value: True, for read-only, otherwise False
        """
        self.setReadOnly(value)

    def set_regex_validator(
        self, regex: str, flags: int = 0
    ) -> gui.RegularExpressionValidator:
        validator = gui.RegularExpressionValidator(self)
        validator.set_regex(regex, flags)
        self.set_validator(validator)
        return validator

    def set_range(self, lower: int | None, upper: int | None):
        val = gui.IntValidator()
        val.set_range(lower, upper)
        self.set_validator(val)

    def set_validator(
        self,
        validator: gui.QValidator
        | ValidatorStr
        | datatypes.PatternType
        | Callable
        | None,
        strict: bool = True,
        empty_allowed: bool | None = None,
        append: bool = False,
        **kwargs,
    ) -> gui.QValidator:
        from prettyqt import validators

        match validator:
            case str() if "|" in validator:
                vals = [get_validator(i, **kwargs) for i in validator.split("|")]
                validator: widgets.QValidator = validators.AndValidator(vals)
            case str() | re.Pattern() | core.QRegularExpression() | Callable():
                validator = get_validator(validator, **kwargs)
            case None | gui.QValidator():
                pass
            case _:
                raise ValueError(validator)
        if empty_allowed is False:
            validator = validators.AndValidator(
                [validator, validators.NotEmptyValidator()]
            )
        elif empty_allowed is True:
            validator = validators.OrValidator([validator, validators.EmptyValidator()])
        if not strict:
            validator = validators.NotStrictValidator(validator)
        if append and (prev := widget.validator()) is not None:
            validator = validators.AndValidator([prev, validator])
        self.setValidator(validator)
        self._update_background()
        return validator

    def set_input_mask(self, mask: str):
        match mask:
            case "ip_address":
                mask = "000.000.000.000;_"
            case "mac_address":
                mask = "HH:HH:HH:HH:HH:HH;_"
            case "iso_date":
                mask = "0000-00-00"
        self.setInputMask(mask)

    def _update_background(self):
        color = None if self.hasAcceptableInput() else self._validation_color
        self.set_background_color(color)

    def set_validation_color(self, color: datatypes.ColorType):
        self._validation_color = colors.get_color(color).as_qt()

    def get_validation_color(self) -> gui.QColor:
        return self._validation_color

    def set_echo_mode(self, mode: EchoModeStr | widgets.QLineEdit.EchoMode):
        """Set echo mode.

        Args:
            mode: echo mode to use
        """
        self.setEchoMode(ECHO_MODE.get_enum_value(mode))

    def get_echo_mode(self) -> EchoModeStr:
        """Return echo mode.

        Returns:
            echo mode
        """
        return ECHO_MODE.inverse[self.echoMode()]

    def set_cursor_move_style(
        self, style: constants.CursorMoveStyleStr | constants.CursorMoveStyle
    ):
        """Set cursor move style.

        Args:
            style: cursor move style to use
        """
        self.setCursorMoveStyle(constants.CURSOR_MOVE_STYLE.get_enum_value(style))

    def get_cursor_move_style(self) -> constants.CursorMoveStyleStr:
        """Return cursor move style.

        Returns:
            cursor move style
        """
        return constants.CURSOR_MOVE_STYLE.inverse[self.cursorMoveStyle()]

    def add_action(
        self,
        action: gui.QAction,
        position: ActionPositionStr | widgets.QLineEdit.ActionPosition = "trailing",
    ):
        self.addAction(action, ACTION_POSITION.get_enum_value(position))

    def set_value(self, value: str):
        self.setText(value)

    def get_value(self) -> str:
        return self.text()

    def is_valid(self) -> bool:
        return self.hasAcceptableInput()


if __name__ == "__main__":
    from prettyqt import debugging

    app = widgets.app()
    widget = LineEdit()

    stalker = debugging.Stalker(widget)
    stalker.hook()
    action = gui.Action(text="hallo", icon="mdi.folder")
    # widget.add_action(action)
    widget.set_validator("website")
    widget.setPlaceholderText("test")
    widget.setClearButtonEnabled(True)
    # widget.set_regex_validator("[0-9]+")
    widget.show()
    with app.debug_mode():
        app.exec()
    print(widget.hasAcceptableInput())



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui, widgets
from prettyqt.qt import QtCharts
from prettyqt.utils import bidict


MarkerShapeStr = Literal["default", "rectangle", "circle", "from_series"]

MARKER_SHAPES: bidict[MarkerShapeStr, QtCharts.QLegend.MarkerShape] = bidict(
    default=QtCharts.QLegend.MarkerShape.MarkerShapeDefault,
    rectangle=QtCharts.QLegend.MarkerShape.MarkerShapeRectangle,
    circle=QtCharts.QLegend.MarkerShape.MarkerShapeCircle,
    from_series=QtCharts.QLegend.MarkerShape.MarkerShapeFromSeries,
)


class Legend(widgets.GraphicsWidgetMixin):
    def __init__(self, item: QtCharts.QLegend):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def set_alignment(self, alignment: constants.SideStr | QtCharts.QLegend.MarkerShape):
        """Set the alignment of the legend.

        Args:
            alignment: alignment for the legend
        """
        self.setAlignment(constants.SIDES.get_enum_value(alignment))

    def get_alignment(self) -> constants.SideStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.SIDES.inverse[self.alignment()]

    def set_marker_shape(self, shape: MarkerShapeStr):
        """Set the marker shape.

        Args:
            shape: marker shape
        """
        self.setMarkerShape(MARKER_SHAPES.get_enum_value(shape))

    def get_marker_shape(self) -> MarkerShapeStr:
        """Return current marker shape.

        Returns:
            Marker shape
        """
        return MARKER_SHAPES.inverse[self.markerShape()]

    def get_border_color(self) -> gui.Color:
        return gui.Color(self.borderColor())

    def get_color(self) -> gui.Color:
        return gui.Color(self.color())

    def get_label_color(self) -> gui.Color:
        return gui.Color(self.labelColor())

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())



from __future__ import annotations

from prettyqt import core, widgets


class RadioButton(widgets.AbstractButtonMixin, widgets.QRadioButton):
    """Radio button with a text label."""

    value_changed = core.Signal(bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.toggled.connect(self.value_changed)


if __name__ == "__main__":
    app = widgets.app()
    widget = RadioButton("This is a test")
    widget.set_icon("mdi.timer")
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterator

from prettyqt import webenginecore


class WebEngineHistory:
    def __init__(self, history: webenginecore.QWebEngineHistory):
        self.history = history

    def __getattr__(self, val):
        return getattr(self.history, val)

    def __len__(self):
        # pyside2 does not support len(self.history)
        return self.history.count()

    def __getitem__(self, index: int) -> webenginecore.WebEngineHistoryItem:
        item = self.history.itemAt(index)
        return webenginecore.WebEngineHistoryItem(item)

    def __iter__(self) -> Iterator[webenginecore.WebEngineHistoryItem]:
        items = [webenginecore.WebEngineHistoryItem(i) for i in self.history.items()]
        return iter(items)

    def get_items(self) -> list[webenginecore.WebEngineHistoryItem]:
        """Get history items."""
        return [webenginecore.WebEngineHistoryItem(i) for i in self.history.items()]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    page = webenginecore.WebEnginePage()
    page.load_url("http://www.google.de")
    page.load_url("http://www.google.com")
    history = page.history()
    item = WebEngineHistory(history)
    print(len(item))
    app.exec()



from __future__ import annotations

from prettyqt import animations, core, eventfilters, widgets


class AnimatedToolTipEventFilter(eventfilters.BaseEventFilter):
    ID = "animated_tooltip"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tool_tip = widgets.Label()
        self.tool_tip.set_flags(frameless=True, tooltip=True)
        self.tool_tip.set_attributes(
            transparent_for_mouse_events=True,
            no_system_background=True,
            translucent_background=True,
        )
        self.tool_tip.hide()
        self.slide_anim = animations.SlideAnimation(self.tool_tip, duration=1000)
        self.slide_anim.apply_to(self.tool_tip)
        self.fade_anim = animations.FadeInAnimation(duration=1000)
        self.fade_anim.apply_to(self.tool_tip)
        self.animation = self.slide_anim | self.fade_anim

    def eventFilter(self, obj: widgets.QWidget, event: core.QEvent) -> bool:
        match event.type():
            case core.QEvent.Type.Enter:
                # self.tool_tip.adjustSize()
                self.tool_tip.setText(obj.toolTip())
                self.tool_tip.show()
                position = "bottom"
                distance = 100
                center = obj.rect().center()
                match position:
                    case "top":
                        center.setY(obj.rect().bottom())
                        delta = core.Point(0, distance)
                    case "bottom":
                        center.setY(obj.rect().top())
                        delta = core.Point(0, -distance)
                    case "left":
                        center.setX(obj.rect().right())
                        delta = core.Point(distance, 0)
                    case "right":
                        center.setX(obj.rect().left())
                        delta = core.Point(-distance, 0)
                center = obj.mapToGlobal(center)
                self.slide_anim.set_start_value(center + delta)
                self.slide_anim.set_end_value(center)
                self.animation.start_animation()
            case core.QEvent.Type.Leave:
                self.tool_tip.hide()
                self.animation.stop()
            case core.QEvent.Type.ToolTip:
                return True
        return super().eventFilter(obj, event)


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.ToolButton()
    ef = AnimatedToolTipEventFilter()
    w.installEventFilter(ef)
    w.setText("test")
    w.set_icon("mdi.timer")
    w.set_tooltip("testus")
    w.show()
    app.exec()



from __future__ import annotations

import logging

from typing import Literal

from prettyqt import core, texttospeech
from prettyqt.utils import bidict


logger = logging.getLogger()

STATE = bidict(
    ready=texttospeech.QTextToSpeech.State.Ready,
    speaking=texttospeech.QTextToSpeech.State.Speaking,
    paused=texttospeech.QTextToSpeech.State.Paused,
    error=texttospeech.QTextToSpeech.State.Error,
)

StateStr = Literal["ready", "speaking", "paused", "error"]


class TextToSpeech(core.ObjectMixin, texttospeech.QTextToSpeech):
    def get_state(self) -> StateStr:
        return STATE.inverse[self.state()]

    def get_locale(self) -> core.Locale:
        return core.Locale(self.locale())

    def get_available_locales(self) -> list[core.Locale]:
        return [core.Locale(locale) for locale in self.availableLocales()]

    def get_voice(self) -> texttospeech.Voice:
        return texttospeech.Voice(self.voice())

    def get_available_voices(self) -> list[texttospeech.Voice]:
        return [texttospeech.Voice(voice) for voice in self.availableVoices()]


if __name__ == "__main__":
    app = core.app()
    print(TextToSpeech.availableEngines())
    val = TextToSpeech("sapi")
    print(val.get_available_voices())
    val.say("Test")
    app.exec()



from __future__ import annotations

import pathlib

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtSpatialAudio
from prettyqt.utils import bidict, datatypes


LoopsStr = Literal["infinite", "once"]

LOOPS: bidict[LoopsStr, QtSpatialAudio.QAmbientSound.Loops] = bidict(
    infinite=QtSpatialAudio.QAmbientSound.Loops.Infinite,
    once=QtSpatialAudio.QAmbientSound.Loops.Once,
)


class AmbientSound(core.ObjectMixin, QtSpatialAudio.QAmbientSound):
    """A stereo overlay sound."""

    def set_source(self, source: datatypes.UrlType):
        self.setSource(datatypes.to_local_url(source))

    def get_source(self) -> pathlib.Path:
        return pathlib.Path(self.source().toString())


if __name__ == "__main__":
    engine = QtSpatialAudio.QAudioEngine()
    sound = AmbientSound(engine)
    sound.get_source()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, widgets
from prettyqt.qt import QtGui, QtWidgets
from prettyqt.utils import bidict


CorrectionModeStr = Literal["to_previous", "to_nearest"]

CORRECTION_MODES: bidict[CorrectionModeStr, QtWidgets.QSpinBox.CorrectionMode] = bidict(
    to_previous=QtWidgets.QSpinBox.CorrectionMode.CorrectToPreviousValue,
    to_nearest=QtWidgets.QSpinBox.CorrectionMode.CorrectToNearestValue,
)

SymbolStr = Literal["up_down", "plus_minus", "none"]

SYMBOLS: bidict[SymbolStr, QtWidgets.QSpinBox.ButtonSymbols] = bidict(
    up_down=QtWidgets.QSpinBox.ButtonSymbols.UpDownArrows,
    plus_minus=QtWidgets.QSpinBox.ButtonSymbols.PlusMinus,
    none=QtWidgets.QSpinBox.ButtonSymbols.NoButtons,
)

StepTypeStr = Literal["default", "adaptive"]

STEP_TYPES: bidict[StepTypeStr, QtWidgets.QSpinBox.StepType] = bidict(
    default=QtWidgets.QSpinBox.StepType.DefaultStepType,
    adaptive=QtWidgets.QSpinBox.StepType.AdaptiveDecimalStepType,
)

StepEnabledFlagStr = Literal["none", "up_enabled", "down_enabled"]

STEP_ENABLED_FLAG: bidict[StepEnabledFlagStr, QtWidgets.QSpinBox.StepEnabledFlag] = (
    bidict(
        none=QtWidgets.QSpinBox.StepEnabledFlag.StepNone,
        up_enabled=QtWidgets.QSpinBox.StepEnabledFlag.StepUpEnabled,
        down_enabled=QtWidgets.QSpinBox.StepEnabledFlag.StepDownEnabled,
    )
)


class AbstractSpinBoxMixin(widgets.WidgetMixin):
    def __init__(self, *args, show_group_separator: bool = True, **kwargs):
        super().__init__(*args, show_group_separator=show_group_separator, **kwargs)
        self.setLineEdit(widgets.LineEdit(parent=self))

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "alignment": constants.ALIGNMENTS,
            "buttonSymbols": SYMBOLS,
            "correctionMode": CORRECTION_MODES,
        }
        return maps

    def is_valid(self) -> bool:
        return self.hasAcceptableInput()

    def set_validator(
        self, validator: QtGui.QValidator | widgets.lineedit.ValidatorStr | None, **kwargs
    ) -> QtGui.QValidator:
        return self.lineEdit().set_validator(validator)

    def get_button_symbols(self) -> SymbolStr:
        """Return button symbol type.

        Returns:
            button symbol type
        """
        return SYMBOLS.inverse[self.buttonSymbols()]

    def set_button_symbols(self, mode: SymbolStr | QtWidgets.QSpinBox.ButtonSymbols):
        """Set button symbol type.

        Args:
            mode: button symbol type to use
        """
        self.setButtonSymbols(SYMBOLS.get_enum_value(mode))

    def set_correction_mode(
        self, mode: CorrectionModeStr | QtWidgets.QSpinBox.CorrectionMode
    ):
        """Set correction mode.

        Args:
            mode: correction mode to use
        """
        self.setCorrectionMode(CORRECTION_MODES.get_enum_value(mode))

    def get_correction_mode(self) -> CorrectionModeStr:
        """Return correction mode.

        Returns:
            correction mode
        """
        return CORRECTION_MODES.inverse[self.correctionMode()]

    def set_step_type(self, mode: StepTypeStr | QtWidgets.QSpinBox.StepType):
        """Set step type.

        Args:
            mode: step type to use
        """
        self.setStepType(STEP_TYPES.get_enum_value(mode))

    def get_step_type(self) -> StepTypeStr:
        """Return step type.

        Returns:
            step type
        """
        return STEP_TYPES.inverse[self.stepType()]

    def set_special_value(self, value: str):
        self.setSpecialValueText(value)

    def get_value(self) -> int:
        return self.value()

    def set_value(self, value: int | float):
        self.setValue(value)


class AbstractSpinBox(AbstractSpinBoxMixin, QtWidgets.QAbstractSpinBox):
    """Spinbox and a line edit to display values."""


if __name__ == "__main__":
    app = widgets.app()
    widget = AbstractSpinBox()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtGui
from prettyqt.utils import datatypes, get_repr


class AbstractTextDocumentLayoutMixin(core.ObjectMixin):
    def __repr__(self):
        return get_repr(self)

    def __len__(self):
        return self.pageCount()

    def get_block_bounding_rect(self, block: QtGui.QTextBlock) -> core.RectF:
        return core.RectF(self.blockBoundingRect(block))

    def get_frame_bounding_rect(self, frame: QtGui.QTextBlock) -> core.RectF:
        return core.RectF(self.frameBoundingRect(frame))

    def hit_test(self, point: datatypes.PointFType, fuzzy: bool = False) -> int | None:
        accuracy = (
            constants.HitTestAccuracy.FuzzyHit
            if fuzzy
            else constants.HitTestAccuracy.ExactHit
        )
        result = self.hitTest(datatypes.to_pointf(point), accuracy)
        return None if result == -1 else result


class AbstractTextDocumentLayout(
    AbstractTextDocumentLayoutMixin, QtGui.QAbstractTextDocumentLayout
):
    """Abstract base class used to implement custom layouts for QTextDocuments."""


if __name__ == "__main__":
    from prettyqt import gui

    doc = gui.TextDocument()
    layout = AbstractTextDocumentLayout(doc)



from __future__ import annotations

import os

from typing import Literal

from prettyqt import core, pdf
from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict, datatypes


ROLE = bidict(
    url=QtWebEngineCore.QWebEngineHistoryModel.Roles.UrlRole,
    title=QtWebEngineCore.QWebEngineHistoryModel.Roles.TitleRole,
    offset=QtWebEngineCore.QWebEngineHistoryModel.Roles.OffsetRole,
    icon_url=QtWebEngineCore.QWebEngineHistoryModel.Roles.IconUrlRole,
)

RoleStr = Literal[
    "url",
    "title",
    "offset",
    "icon_url",
]


class WebEngineHistoryModel(
    core.AbstractListModelMixin, QtWebEngineCore.QWebEngineHistoryModel
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setDocument(pdf.PdfDocument(self))

    def set_document(self, document: datatypes.PathType | QtWebEngineCore.QPdfDocument):
        if not isinstance(document, QtWebEngineCore.QPdfDocument):
            path = os.fspath(document)
            document = pdf.PdfDocument(self)
            document.load(path)
        self.setDocument(document)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    # model = WebEngineHistoryModel()
    widget = widgets.TableView()
    # widget.set_model(model)
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterator
import contextlib

from typing import Literal

from prettyqt import constants, core, gui
from prettyqt.utils import bidict, colors, datatypes


CompositionModeStr = Literal[
    "source_over",
    "destination_over",
    "clear",
    "source",
    "destination",
    "source_in",
    "destination_in",
    "source_out",
    "destination_out",
    "source_atop",
    "destination_atop",
]

COMPOSITION_MODE: bidict[CompositionModeStr, gui.QPainter.CompositionMode] = bidict(
    source_over=gui.QPainter.CompositionMode.CompositionMode_SourceOver,
    destination_over=gui.QPainter.CompositionMode.CompositionMode_DestinationOver,
    clear=gui.QPainter.CompositionMode.CompositionMode_Clear,
    source=gui.QPainter.CompositionMode.CompositionMode_Source,
    destination=gui.QPainter.CompositionMode.CompositionMode_Destination,
    source_in=gui.QPainter.CompositionMode.CompositionMode_SourceIn,
    destination_in=gui.QPainter.CompositionMode.CompositionMode_DestinationIn,
    source_out=gui.QPainter.CompositionMode.CompositionMode_SourceOut,
    destination_out=gui.QPainter.CompositionMode.CompositionMode_DestinationOut,
    source_atop=gui.QPainter.CompositionMode.CompositionMode_SourceAtop,
    destination_atop=gui.QPainter.CompositionMode.CompositionMode_DestinationAtop,
)

RenderHintStr = Literal[
    "antialiasing",
    "text_antialiasing",
    "smooth_pixmap_transform",
    "lossless_image_rendering",
]

RENDER_HINTS: bidict[RenderHintStr, gui.QPainter.RenderHint] = bidict(
    antialiasing=gui.QPainter.RenderHint.Antialiasing,
    text_antialiasing=gui.QPainter.RenderHint.TextAntialiasing,
    smooth_pixmap_transform=gui.QPainter.RenderHint.SmoothPixmapTransform,
    lossless_image_rendering=gui.QPainter.RenderHint.LosslessImageRendering,
)


class PainterMixin:
    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        self.end()

    @contextlib.contextmanager
    def paint_on(self, obj) -> Iterator[Painter]:
        self.begin(obj)
        yield self
        self.end()

    @contextlib.contextmanager
    def backup_state(self) -> Iterator[Painter]:
        self.save()
        yield self
        self.restore()

    @contextlib.contextmanager
    def native_mode(self) -> Iterator[Painter]:
        self.beginNativePainting()
        yield self
        self.endNativePainting()

    @contextlib.contextmanager
    def edit_pen(self) -> Iterator[gui.Pen]:
        pen = gui.Pen(self.pen())
        yield pen
        self.setPen(pen)

    @contextlib.contextmanager
    def edit_font(self) -> Iterator[gui.Font]:
        font = gui.Font(self.font())
        yield font
        self.setFont(font)

    def draw_text(
        self,
        position: datatypes.PointType | datatypes.RectType | datatypes.RectFType,
        text: str,
        alignment: constants.AlignmentStr = "center",
    ):
        match position:
            case (_, _):
                position = core.Point(*position)
            case (_, _, _, _):
                position = core.RectF(*position)
        self.drawText(position, constants.ALIGNMENTS[alignment].value, text)

    def draw_image(
        self,
        target: core.QPoint | core.QPointF | core.QRect | core.QRectF,
        frame_buffer: gui.QImage,
    ):
        self.set_composition_mode("source_atop")
        self.drawImage(target, frame_buffer)

    def draw_polygon(
        self,
        points: (gui.QPolygon | gui.QPolygonF | list[core.QPoint] | list[core.QPointF]),
        fill_rule: constants.FillRuleStr | constants.FillRule = "odd_even",
    ):
        self.drawPolygon(points, fillRule=constants.FILL_RULE.get_enum_value(fill_rule))

    def draw_rounded_rect(
        self,
        rect: datatypes.RectType | datatypes.RectFType,
        x_radius: float,
        y_radius: float,
        relative: bool = False,
    ):
        flag = (
            constants.SizeMode.RelativeSize
            if relative
            else constants.SizeMode.AbsoluteSize
        )
        self.drawRoundedRect(datatypes.to_rect(rect), x_radius, y_radius, flag)

    def draw_star(self, size: float = 1.0, fill_rule: constants.FillRuleStr = "winding"):
        star = gui.PolygonF.create_star(size)
        self.drawPolygon(star, constants.FILL_RULE[fill_rule])

    def draw_diamond(
        self, size: float = 1.0, fill_rule: constants.FillRuleStr = "winding"
    ):
        star = gui.PolygonF.create_diamond(size)
        self.drawPolygon(star, constants.FILL_RULE[fill_rule])

    def use_antialiasing(self):
        self.setRenderHint(self.RenderHint.Antialiasing, True)

    def fill_rect(
        self,
        rect: datatypes.RectType | datatypes.RectFType,
        color: datatypes.ColorType,
        pattern: constants.BrushStyleStr | constants.BrushStyle = "solid",
    ):
        color = colors.get_color(color)
        if pattern != "solid":
            color = gui.Brush(color, constants.BRUSH_STYLE.get_enum_value(pattern))
        self.fillRect(datatypes.to_rect(rect), color)

    def set_pen(
        self,
        style: constants.PenStyleStr | None = "solid",
        width: float = 1.0,
        color: datatypes.ColorType = "black",
        brush: gui.QBrush | None = None,
        miter_limit: float = 2.0,
        join_style: constants.JoinStyleStr = "bevel",
        cap_style: constants.CapStyleStr = "square",
    ) -> gui.Pen:
        """Set pen to use.

        Args:
            style: pen style
            width: pen width
            color: pen color
            brush: pen brush
            miter_limit: miter limit
            join_style: pen join style
            cap_style: pen cap style
        """
        pen = gui.Pen()
        pen.set_style(style)
        if style in ["none", None]:
            self.setPen(pen)
            return pen
        pen.set_cap_style(cap_style)
        pen.set_join_style(join_style)
        pen.setMiterLimit(miter_limit)
        pen.setWidthF(width)
        if brush is not None:
            pen.setBrush(brush)
        pen.set_color(color)
        self.setPen(pen)
        return pen

    def get_pen(self) -> gui.Pen:
        """Return current pen.

        Returns:
            current pen
        """
        return gui.Pen(self.pen())

    def set_color(self, color: datatypes.ColorType):
        color = colors.get_color(color)
        self.setPen(color)

    def set_brush(self, brush: gui.QBrush | datatypes.ColorType):
        if not isinstance(brush, gui.QBrush):
            brush = colors.get_color(brush)
        self.setBrush(brush)

    def set_transparent_background(self, transparent: bool = True):
        mode = (
            constants.BGMode.TransparentMode
            if transparent
            else constants.BGMode.OpaqueMode
        )
        self.setBackgroundMode(mode)

    def set_composition_mode(
        self, mode: CompositionModeStr | gui.QPainter.CompositionMode
    ):
        """Set the current composition mode.

        Arguments:
            mode: composition mode
        """
        self.setCompositionMode(COMPOSITION_MODE.get_enum_value(mode))

    def get_composition_mode(self) -> CompositionModeStr:
        """Get the current composition mode.

        Returns:
            composition mode
        """
        return COMPOSITION_MODE.inverse[self.compositionMode()]

    def set_transform(self, transform: datatypes.TransformType, combine: bool = False):
        self.setTransform(datatypes.to_transform(transform), combine)

    def get_font_metrics(self) -> gui.FontMetrics:
        return gui.FontMetrics(self.fontMetrics())

    def set_clip_path(
        self,
        path: gui.QPainterPath,
        operation: constants.ClipOperationStr | constants.ClipOperation = "replace",
    ):
        self.setClipPath(path, constants.CLIP_OPERATION.get_enum_value(operation))

    def get_text_rect(self, text: str) -> core.Rect:
        return self.drawText(
            core.Rect(), constants.TextFlag.TextDontPrint, text
        )  # type: ignore

    @contextlib.contextmanager
    def clip_path(
        self, operation: constants.ClipOperationStr = "replace"
    ) -> Iterator[gui.PainterPath]:
        path = gui.PainterPath()
        yield path
        self.set_clip_path(path, operation)

    @contextlib.contextmanager
    def apply_transform(self, combine: bool = True) -> Iterator[gui.Transform]:
        transform = gui.Transform()
        yield transform
        self.setTransform(transform, combine)

    @contextlib.contextmanager
    def offset_by(self, x: int = 0, y: int = 0) -> Iterator[Painter]:
        self.translate(x, y)
        yield self
        self.translate(-x, -y)


class Painter(PainterMixin, gui.QPainter):
    """Performs low-level painting on widgets and other paint devices."""


if __name__ == "__main__":
    painter = Painter()
    painter.draw_text((0, 0, 1, 1), "aaa")



from __future__ import annotations

import json

from prettyqt import gui


class JsonValidator(gui.Validator):
    """Validator which checks for strings which can be parsed as JSON."""

    ID = "json"

    def __eq__(self, other: object):
        return isinstance(other, JsonValidator)

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> tuple[gui.QValidator.State, str, int]:
        try:
            json.loads(text)
            return self.State.Acceptable, text, pos
        except json.decoder.JSONDecodeError:
            return self.State.Intermediate, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = JsonValidator()
    app = widgets.app()
    widget = widgets.LineEdit()
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, eventfilters


if TYPE_CHECKING:
    from prettyqt import widgets


class ListViewGridResizeEventFilter(eventfilters.BaseEventFilter):
    """Eventfilter that scales grid size of ListViews."""

    ID = "listview_grid_resize"

    def __init__(
        self, num_columns: int = 5, parent: widgets.ListView | None = None, **kwargs
    ):
        super().__init__(parent=parent, **kwargs)
        self._view_columns = num_columns

    def eventFilter(self, source, event: core.Event) -> bool:
        match event.type():
            case core.Event.Type.Resize:
                self._resize(source)
        return super().eventFilter(source, event)

    def _resize(self, source: widgets.ListView):
        width = source.width() - 30
        # The minus 30 above ensures we don't end up with an item width that
        # can't be drawn the expected number of times across the view without
        # being wrapped. Without this, the view can flicker during resize
        tile_width = int(width / self._view_columns)
        icon_width = int(tile_width * 0.8)
        source.setGridSize(core.QSize(tile_width, tile_width))
        source.setIconSize(core.QSize(icon_width, icon_width))


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.ListView()
    # widget.set_icon("mdi.folder")
    test = ListViewGridResizeEventFilter(parent=widget)
    widget.installEventFilter(test)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Any

from prettyqt import core
from prettyqt.utils import get_repr


class RegularExpressionMatch(core.QRegularExpressionMatch):
    """The results of a matching a QRegularExpression against a string."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.string = None
        self.pos = None
        self.endpos = None

    def __repr__(self):
        return get_repr(self)

    def __getitem__(self, item: int | str) -> str:
        if captured := self.captured(item):
            return captured
        else:
            raise KeyError(item)

    def __bool__(self):
        return self.isValid()

    def get_match_type(self) -> core.regularexpression.MatchTypeStr:
        return core.regularexpression.MATCH_TYPE.inverse[self.matchType()]

    def group(self, *groups: int | str) -> tuple[str, ...] | str:
        match len(groups):
            case 0:
                return self.captured(0)
            case 1:
                return self.captured(groups[0])
            case _:
                return tuple(self.captured(i) for i in groups)

    def groups(self, default=None) -> tuple:
        if self.lastindex is None:
            return ()
        return tuple(
            self.group(i) if i <= self.lastindex else default
            for i in range(self.re.captureCount())
        )

    def groupdict(self, default=None) -> dict[str, Any]:
        if self.lastindex is None:
            return {}
        groups = [
            self.group(i) if i <= self.lastindex else default
            for i in range(self.re.captureCount())
        ]
        names = self.re.namedCaptureGroups()
        return {names[i]: groups[i] for i in range(self.re.captureCount())}

    def start(self, group: int = 0) -> int:
        return self.capturedStart(group)

    def end(self, group: int = 0) -> int:
        return self.capturedEnd(group)

    def span(self, group: int = 0) -> tuple[int, int]:
        return (self.capturedStart(group), self.capturedEnd(group))

    @property
    def lastindex(self) -> int | None:
        idx = self.lastCapturedIndex()
        return None if idx == -1 else idx

    @property
    def lastgroup(self) -> str | None:
        if self.lastCapturedIndex() == -1:
            return None
        return self.re.namedCaptureGroups()[self.lastCapturedIndex()]

    @property
    def re(self) -> core.QRegularExpression:
        return self.regularExpression()

    @property
    def partial(self) -> bool:
        return self.hasPartialMatch()


if __name__ == "__main__":
    reg = RegularExpressionMatch()



from __future__ import annotations

from prettyqt.qt.QtMultimediaWidgets import *  # noqa: F403

from .graphicsvideoitem import GraphicsVideoItem
from .videowidget import VideoWidget
from prettyqt.qt import QtMultimediaWidgets

QT_MODULE = QtMultimediaWidgets

__all__ = ["VideoWidget", "GraphicsVideoItem"]



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict, serializemixin


NamedColorSpaceStr = Literal[
    "srgb", "srgb_linear", "adobe_rgb", "display_p3", "pro_photo_rgb"
]

NAMED_COLOR_SPACE: bidict[NamedColorSpaceStr, QtGui.QColorSpace.NamedColorSpace] = bidict(
    srgb=QtGui.QColorSpace.NamedColorSpace.SRgb,
    srgb_linear=QtGui.QColorSpace.NamedColorSpace.SRgbLinear,
    adobe_rgb=QtGui.QColorSpace.NamedColorSpace.AdobeRgb,
    display_p3=QtGui.QColorSpace.NamedColorSpace.DisplayP3,
    pro_photo_rgb=QtGui.QColorSpace.NamedColorSpace.ProPhotoRgb,
)

PrimariesStr = Literal["custom", "srgb", "adobe_rgb", "dci_p3_d65", "pro_photo_rgb"]

PRIMARIES: bidict[PrimariesStr, QtGui.QColorSpace.Primaries] = bidict(
    custom=QtGui.QColorSpace.Primaries.Custom,
    srgb=QtGui.QColorSpace.Primaries.SRgb,
    adobe_rgb=QtGui.QColorSpace.Primaries.AdobeRgb,
    dci_p3_d65=QtGui.QColorSpace.Primaries.DciP3D65,
    pro_photo_rgb=QtGui.QColorSpace.Primaries.ProPhotoRgb,
)


TransformFunctionStr = Literal["custom", "linear", "gamma", "srgb", "pro_photo_rgb"]

TRANSFER_FUNCTION: bidict[TransformFunctionStr, QtGui.QColorSpace.TransferFunction] = (
    bidict(
        custom=QtGui.QColorSpace.TransferFunction.Custom,
        linear=QtGui.QColorSpace.TransferFunction.Linear,
        gamma=QtGui.QColorSpace.TransferFunction.Gamma,
        srgb=QtGui.QColorSpace.TransferFunction.SRgb,
        pro_photo_rgb=QtGui.QColorSpace.TransferFunction.ProPhotoRgb,
    )
)


class ColorSpace(serializemixin.SerializeMixin, QtGui.QColorSpace):
    """Color space abstraction."""

    def __bool__(self):
        return self.isValid()

    def set_primaries(self, primaries: PrimariesStr | QtGui.QColorSpace.Primaries):
        """Set primaries.

        Args:
            primaries: primaries to use
        """
        self.setPrimaries(PRIMARIES.get_enum_value(primaries))

    def get_primaries(self) -> PrimariesStr:
        """Return current primaries.

        Returns:
            primaries
        """
        return PRIMARIES.inverse[self.primaries()]

    def set_transfer_function(
        self,
        fn: TransformFunctionStr | QtGui.QColorSpace.TransferFunction,
        gamma: float = 0.0,
    ):
        """Set transfer function.

        Args:
            fn: transfer function to use
            gamma: gamma value
        """
        self.setTransferFunction(TRANSFER_FUNCTION.get_enum_value(fn), gamma)

    def get_transfer_function(self) -> TransformFunctionStr:
        """Return current transfer function.

        Returns:
            transfer function
        """
        return TRANSFER_FUNCTION.inverse[self.transferFunction()]


if __name__ == "__main__":
    space = ColorSpace()



from __future__ import annotations

from collections.abc import Sequence
import logging

import attrs

from prettyqt import itemmodels
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class AttrsModel(itemmodels.BaseDataclassModel):
    """Table model to display a list of [attrs](https://www.attrs.org/) dataclasses."""

    SUPPORTS = Sequence[datatypes.IsAttrs]

    def __init__(self, items: Sequence[datatypes.IsAttrs], **kwargs):
        super().__init__(items, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (datatypes.IsAttrs(), *_):
                return True
            case _:
                return False

    def get_fields(self):
        fields = attrs.fields(self.Class)
        return {field.name: field for field in fields}


if __name__ == "__main__":
    from prettyqt import widgets

    @attrs.define
    class BaseSetting:
        name: str
        label: str
        description: str
        requires_restart: bool = False

    @attrs.define(frozen=True)
    class SelectionSetting(BaseSetting):
        options: list[str] | None = None
        minimum: float = 0.0
        maximum: float = 0.0
        pattern: str = ""
        requires_restart: bool = False

    app_style = SelectionSetting(
        name="app_style",
        label="App style",
        description="Some longer text",
        options=[],
        requires_restart=True,
    )
    app = widgets.app()
    items = [BaseSetting(name="test", label="some label", description="jhaj"), app_style]
    klasses = [type(i) for i in items]
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    model = AttrsModel(items, parent=view)
    view.set_model(model)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate("editor")
    view.resize(1000, 1000)
    with app.debug_mode():
        view.show()
        app.exec()



"""Package containing stuff related to animating widgets.


```py
# a lot of animations depend on current geometry.
# It´s advisable to show widget first so that it gets layouted.
widget = widget.Label("I will get animated.")
widget.show()

# start a general fade animation with fixed start/end
widget.fx["windowOpacity"].animate(start=0, end=1, duration=1000)

# Transition widget size from current value to (500, 500)
widget.fx["size"].transition_to((500, 500), duration=1000)

# Transition widget size from current value to (500, 500)
widget.fx["pos"].transition_from((0, -100), duration=1000)

# fade-in when widget gets clicked
widget.fx["windowOpacity"].animate_on_event("mouse_button_press", start=0, end=1)

# convenience method for mentioned fade-in
widget.fx.fade_in(start=0, end=1, start=500)

# start a zoom animation
widget.fx.zoom(start=1, end=1.5, duration=400)

# start a slide animation 100px to the right
widget.fx.slide(start=(0, 0), end=(100, 0), duration=400)
```

PrettyQt makes it easy to animate your widgets.
All animation-related methods are available via the Widget.fx attribute.

Animating stylesheet values is also possible easily.

TODO.


"""

from .cursormoveanimation import CursorMoveAnimation
from .slideanimation import SlideAnimation
from .textanimation import TextAnimation
from .zoomanimation import ZoomAnimation
from .fx import Fx
from .animator import Animator


__all__ = [
    "SlideAnimation",
    "ZoomAnimation",
    "TextAnimation",
    "CursorMoveAnimation",
    "Animator",
    "Fx",
]



# modified version of https://github.com/nicoddemus/qt-async-threads/

from __future__ import annotations

from collections.abc import (
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Generator,
    Iterable,
    Iterator,
)
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import suppress
import dataclasses
import functools
import itertools
import logging
import os
import threading
import time

from typing import Any, ParamSpec, TypeVar, cast

from prettyqt import core


log = logging.getLogger(__name__)

Params = ParamSpec("Params")
T = TypeVar("T")


def chunked_iter(src: Iterable, size: int):
    if not src:
        return
    src_iter = iter(src)
    while cur_chunk := list(itertools.islice(src_iter, size)):
        yield cur_chunk


class AsyncRunner:
    """A runner which runs long-lasting functions using a thread pool."""

    def __init__(self, max_threads: int | None = None):
        super().__init__()
        self._max_threads = max_threads or os.cpu_count() or 1
        self._pool = ThreadPoolExecutor(max_workers=max_threads)

        # Keep track of running tasks,
        # mostly to know if we are idle or not.
        self._running_tasks: list[_AsyncTask] = []

        # This signaller object is used to signal to us when a future running
        # in another thread finishes. Thanks to Qt's queued connections,
        # signals can be safely emitted from separate threads and are queued
        # to run in the same thread as the object receiving it lives (the main
        # thread in our case).
        self._signaller = _FutureDoneSignaller()
        self._signaller.future_done_signal.connect(self._resume_coroutine)

    def __enter__(self: T) -> T:
        return self

    def __exit__(self, *exc_info: object):
        self.close()

    def to_sync(
        self, async_func: Callable[..., Coroutine[Any, Any, None]]
    ) -> Callable[..., None]:
        """Return new sync function.

        Return a new sync function that will start its coroutine using
        ``start_coroutine`` when called, returning immediately.
        Use to connect Qt signals to async functions

        Arguments:
            async_func: Async Qt Slot

        Returns:
            new sync function

        ### Example:
        ```py
        dl_btn.clicked.connect(self.runner.to_sync(self._on_dl_btn_clicked)
        ...
        search_response = await self.runner.run(requests.get, api_url)
        ```
        """

        @functools.wraps(async_func)
        def func(*args: object, **kwargs: object):
            gen = async_func(*args, **kwargs)
            self.start_coroutine(gen)

        return func

    @property
    def max_threads(self) -> int:
        """Return the maximum number of threads used by the internal threading pool."""
        return self._max_threads

    def is_idle(self) -> bool:
        return len(self._running_tasks) == 0

    def close(self):
        self._pool.shutdown(wait=True, cancel_futures=True)

    async def run(
        self, func: Callable[Params, T], *args: Params.args, **kwargs: Params.kwargs
    ) -> T:
        """Run the given function in a thread.

        While it is running, yields
        the control back to the Qt event loop.

        When the thread finishes, this async function resumes, returning
        the return value from the function.

        Arguments:
            func: Callable to run in a thrad
            args: arguments passed to Callable
            kwargs: keyword arguments passed to Callable

        Returns:
            return value from function
        """
        funcs = [functools.partial(func, *args, **kwargs)]
        async for result in self.run_parallel(funcs):
            return result

    async def run_parallel(  # type:ignore[override]
        self, funcs: Iterable[Callable[[], T]]
    ) -> AsyncIterator[T]:
        """Run functions in parallel.

        Run functions in parallel without arguments, yielding results as they get ready.
        (use ``functools.partial`` as necessary)

        Arguments:
            funcs: Sequence of Functions to call.
        """
        # We submit the functions in batches to avoid overloading the pool,
        # which can cause other coroutines to stall.
        # For example, Autofit works by executing 1000s of small functions,
        # which might take a few ms each. If we submitted all those 1000s of functions
        # at once (at the time ``run_parallel`` is called), then other coroutines
        # that try to submit functions to execute in threads would only be resumed
        # much later, causing a noticeable slow down in the application.
        batch_size = max(self._max_threads // 2, 1)
        for function_batch in chunked_iter(funcs, batch_size):
            # Submit all functions from the current batch to the thread pool,
            # using the _AsyncTask to track the futures and await when they finish.
            task = _AsyncTask({self._pool.submit(f) for f in function_batch})
            self._running_tasks.append(task)
            for future in task.futures:
                future.add_done_callback(
                    functools.partial(self._on_task_future_done, task=task)
                )
            try:
                # We wait until all functions in this batch have finished
                # until we submit the next batch. This is not 100% optimal
                # but ensures we get a fairer share of the pool.
                while task.futures:
                    await task
                    for result in task.pop_done_futures():
                        yield result
            finally:
                task.shutdown()
                self._running_tasks.remove(task)

    def _on_future_done(
        self,
        future: Future,
        *,
        coroutine: Coroutine[Any, Any, Any],
    ):
        """Called when a ``Future`` that was submitted to the thread pool finishes.

        This function is called from a separate thread, so we emit the signal
        of the internal ``_signaller``, which thanks to Qt's queued connections feature,
        will post the event to the main loop, and it will be processed there.

        Arguments:
            future: future which finished
            coroutine: Coroutine
        """
        self._signaller.future_done_signal.emit(future, coroutine)

    def _on_task_future_done(self, future: Future, *, task: _AsyncTask):
        """Called when a ``Future`` belonging to a ``_AsyncTask`` has finished.

        Similar to ``_on_future_done``, this emits a signal so the coroutine is resumed
        in the main thread.

        Arguments:
            future: future which finished
            task:  Async Task
        """
        # At this point, we want to get the coroutine associated with the task, and resume
        # its execution in the main loop, but we must take care bc this callback is called
        # from another thread, and it may be called multiple times in succession, but we
        # want only **one** event to be sent, that's
        # why we use ``pop_coroutine``, which is lock-protected and will return
        # the coroutine and set it to None, so next calls of this method will get ``None``
        # and won't trigger the event again.
        if coroutine := task.pop_coroutine():
            self._signaller.future_done_signal.emit(future, coroutine)

    def _resume_coroutine(
        self,
        future: Future,
        coroutine: Coroutine[Any, Any, Any],
    ):
        """Resume paused coroutine.

        Slots connected to our internal ``_signaller`` object,
        called in the main thread after a future finishes, resuming the paused coroutine.

        Arguments:
            future: not used
            coroutine: coroutine to resume
        """
        assert threading.current_thread() is threading.main_thread()
        self.start_coroutine(coroutine)

    def start_coroutine(self, coroutine: Coroutine):
        """Start the coroutine, and returns immediately.

        Arguments:
            coroutine: coroutine to start
        """
        # Note: this function will also be called to resume a paused coroutine that was
        # waiting for a thread to finish (by ``_resume_coroutine``).
        with suppress(StopIteration):
            value = coroutine.send(None)
            if isinstance(value, _AsyncTask):
                # At this point, return control to the event loop; when one
                # of the futures running in the task finishes, it will resume the
                # coroutine back in the main thread.
                value.coroutine = coroutine
                return
            raise ValueError(value)

    def run_coroutine(self, coroutine: Coroutine[Any, Any, T]) -> T:
        """Start the coroutine.

        Starts the coroutine, doing a busy loop while waiting for it to complete,
        returning then the result.

        Arguments:
            coroutine: coroutine to start

        Note: see warning in AbstractAsyncRunner about when to use this function.
        """
        result: T | None = None
        exception: Exception | None = None
        completed = False

        async def wrapper():
            nonlocal result, exception, completed
            try:
                result = await coroutine
            except Exception as e:
                exception = e
            completed = True

        self.start_coroutine(wrapper())
        while not completed:
            core.CoreApplication.processEvents()
            time.sleep(0.01)

        if exception is not None:
            raise exception
        return cast(T, result)


@dataclasses.dataclass
class _AsyncTask(Awaitable[None]):
    """Awaitable that is propagated up the async stack, containing running futures.

    It "awaits" while all futures are processing, and will
    stop and return once any of them complete.
    """

    futures: set[Future]
    coroutine: Coroutine | None = None
    _lock: threading.Lock = dataclasses.field(default_factory=threading.Lock)

    def __await__(self) -> Generator[Any, None, None]:
        if any(x for x in self.futures if x.done()):
            return None
        else:
            yield self

    def pop_done_futures(self) -> Iterator[Any]:
        """Yield done futures, removing them from the ``futures`` simultaniously."""
        for future in list(self.futures):
            if future.done():
                self.futures.discard(future)
                yield future.result()

    def pop_coroutine(self) -> Coroutine | None:
        """Return current coroutine.

        Returns the current coroutine associated with this object, while
        also setting the ``coroutine`` attribute to ``None``. This
        is meant to be called from multiple threads, in a way that only
        one of them will be able to obtain the coroutine object, while the
        others will get ``None``.
        """
        with self._lock:
            coroutine = self.coroutine
            self.coroutine = None
            return coroutine

    def shutdown(self):
        """Cancel any running futures and clears up its attributes."""
        msg = (
            "Should always be called from ``run_parallel``, "
            "at which point this should not have a coroutine associated"
        )
        assert self.coroutine is None, msg
        while self.futures:
            future = self.futures.pop()
            future.cancel()


class _FutureDoneSignaller(core.Object):
    """Used to emit a signal to the main thread when a future finishes in another."""

    # This emits(Future, Coroutine). We need to use ``object`` as the
    # second parameter because Qt doesn't allow us to use an ABC class there it seems.
    future_done_signal = core.Signal(Future, object)


if __name__ == "__main__":
    from pathlib import Path
    from urllib.parse import urlsplit

    import requests

    from requests import Response
    from requests.exceptions import ConnectionError

    from prettyqt import widgets

    class Window(widgets.Widget):
        def __init__(self, directory: Path):
            super().__init__()

            self.setWindowTitle("Cat Downloader")
            self.directory = directory
            self._cancelled = False

            # Build controls.
            self.count_spin = widgets.SpinBox(value=5, minimum=1)
            self.progress_label = widgets.Label("Idle, click below to start downloading")
            self.download_button = widgets.PushButton(
                "Download", clicked=self.on_download_button_clicked
            )
            self.stop_button = widgets.PushButton(
                "Stop", enabled=False, clicked=self.on_cancel_button_clicked
            )

            layout = widgets.FormLayout(self)
            layout.addRow("How many cats?", self.count_spin)
            layout.addRow("Status", self.progress_label)
            layout.addRow(self.download_button)
            layout.addRow(self.stop_button)

        async def on_download_button_clicked(self, checked: bool = False):
            self.progress_label.setText("Searching...")
            self.download_button.setEnabled(False)
            self.stop_button.setEnabled(True)

            self._cancelled = False
            downloaded_count = 0

            def download_one() -> Response:
                # Search.
                search_response = requests.get(
                    "https://api.thecatapi.com/v1/images/search"
                )
                search_response.raise_for_status()

                # Download.
                url = search_response.json()[0]["url"]
                return requests.get(url)

            try:
                functions = [download_one for _ in range(self.count_spin.value())]
                async for response in widgets.app().run_parallel(functions):
                    # Save the contents of the image to a file.
                    parts = urlsplit(response.url)
                    path = (
                        self.directory
                        / f"{downloaded_count:02d}_cat{Path(parts.path).suffix}"
                    )
                    path.write_bytes(response.content)
                    downloaded_count += 1

                    # Show progress.
                    self.progress_label.setText(f"Downloaded {path.name}")
                    widgets.app().processEvents()
                    if self._cancelled:
                        widgets.MessageBox.information(
                            self, "Cancelled", "Download cancelled"
                        )
                        return
            except ConnectionError as e:
                widgets.MessageBox.critical(
                    self, "Error", f"Error connecting to TheCatApi:\n{e}"
                )
                return
            finally:
                self.progress_label.setText(f"Done, {downloaded_count} cats downloaded")
                self.download_button.setEnabled(True)
                self.stop_button.setEnabled(False)

        def on_cancel_button_clicked(self):
            self._cancelled = True

    app = widgets.app()
    win = Window(Path(__file__).parent)
    win.show()
    app.exec()



from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Literal

from prettyqt import constants, core, gui
from prettyqt.utils import serializemixin


if TYPE_CHECKING:
    from prettyqt import widgets


logger = logging.getLogger(__name__)


class Cursor(serializemixin.SerializeMixin, gui.QCursor):
    """Provides a mouse cursor with an arbitrary shape.

    This class is mainly used to create mouse cursors that are associated
    with particular widgets and to get and set the position of the mouse cursor.
    Qt has a number of standard cursor shapes, but you can also make custom cursor
    shapes based on a QBitmap, a mask and a hotspot.
    To associate a cursor with a widget, use `widget.set_cursor()`.
    To associate a cursor with all widgets (normally for a short period of time),
    use `GuiApplication.set_override_cursor()`.
    To set a cursor shape use `QCursor.setShape()` or use the QCursor
    constructor which takes the shape as argument, or you can use one of the
    predefined cursors defined in the `Qt.CursorShape` enum.
    If you want to create a cursor with your own bitmap,
    either use the QCursor constructor which takes a bitmap and a mask
    or the constructor which takes a pixmap as arguments.
    To set or get the position of the mouse cursor use the
    static methods `QCursor.pos()` and `QCursor.setPos()`.

    !!! note:
        It is possible to create a QCursor before QGuiApplication,
        but it is not useful except as a place-holder for a real QCursor
        created after QGuiApplication.
        Attempting to use a QCursor that was created before QGuiApplication will
        result in a crash.

    """

    @classmethod
    def fake_mouse_move(cls):
        cls.setPos(cls.pos() + core.Point(0, 1))
        gui.GuiApplication.processEvents()
        cls.setPos(cls.pos() - core.Point(0, 1))

    @classmethod
    def click(cls, key=constants.MouseButton.LeftButton):
        from prettyqt import widgets

        app = widgets.app()
        widget = app.widgetAt(cls.pos())
        if widget is None:
            return
        pos = cls.pos().toPointF()
        local = pos - widget.mapToGlobal(core.PointF(0, 0))
        logger.info(f"sending MouseClick events to {widget} at {local}")
        event = gui.QMouseEvent(
            core.QEvent.Type.MouseButtonPress,
            local,
            pos,
            key,
            constants.MouseButton.NoButton,
            constants.KeyboardModifier(0),
        )

        core.CoreApplication.sendEvent(widget, event)
        event = gui.QMouseEvent(
            core.QEvent.Type.MouseButtonRelease,
            local,
            pos,
            key,
            constants.MouseButton.NoButton,
            constants.KeyboardModifier(0),
        )

        core.CoreApplication.sendEvent(widget, event)

    def set_shape(self, shape: constants.CursorShapeStr):
        """Set cursor shape.

        Args:
            shape: shape to use
        """
        self.setShape(constants.CURSOR_SHAPE.get_enum_value(shape))

    def get_shape(self) -> constants.CursorShapeStr:
        """Return current cursor shape.

        Returns:
            cursor shape
        """
        return constants.CURSOR_SHAPE.inverse[self.shape()]

    @classmethod
    def get_pos(cls) -> core.Point:
        return core.Point(cls.pos())

    @classmethod
    def set_pos(
        cls,
        where: Literal["screen", "current"]
        | gui.QScreen
        | widgets.QWidget
        | core.QRect
        | core.QPoint
        | tuple[int, int]
        | tuple[int, int, int, int],
        how: Literal[
            "center",
            "top",
            "left",
            "bottom",
            "right",
            "top_left",
            "top_right",
            "bottom_left",
            "bottom_right",
        ] = "center",
        x_offset: int = 0,
        y_offset: int = 0,
        duration: int = 0,
    ):
        """Position cursor onto screen position / widget / window / screen.

        Arguments:
            where: where to position on
            how: How to align
            x_offset: additional x offset for final position
            y_offset: additional y offset for final position
            duration: movement time
        """
        match where:
            case "current":
                p = cls.pos()
                geom = core.Rect(p, p)
            case core.QPoint():
                geom = core.Rect(where, where)
            case (int(), int()):
                p = core.Point(*where)
                geom = core.Rect(p, p)
            case (int(), int(), int(), int()):
                geom = core.Rect(*where)
            case core.QRect():
                geom = where
            case "screen":
                geom = gui.GuiApplication.primaryScreen().geometry()
            case gui.QScreen():
                geom = where.geometry()
            case _ if hasattr(where, "frameGeometry"):  # avoiding to import widgets here
                geom = where.frameGeometry()
            case _:
                raise TypeError(where)
        match how:
            case "center":
                new = geom.center()
            case "top":
                new = core.Point(geom.center().x(), geom.top())
            case "bottom":
                new = core.Point(geom.center().x(), geom.bottom())
            case "left":
                new = core.Point(geom.left(), geom.center().y())
            case "right":
                new = core.Point(geom.right(), geom.center().y())
            case "top_right":
                new = geom.topRight()
            case "top_left":
                new = geom.topLeft()
            case "bottom_right":
                new = geom.bottomRight()
            case "bottom_left":
                new = geom.bottomLeft()
            case _:
                raise TypeError(how)
        new_pos = core.Point(new.x() + x_offset, new.y() + y_offset)
        if duration > 0:
            from prettyqt.animations import cursormoveanimation

            cls._cursor_animation = cursormoveanimation.CursorMoveAnimation(
                duration=duration, end=new_pos
            )
            cls._cursor_animation.start()
        else:
            cls.setPos(new_pos)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.PushButton()
    widget.show()
    with app.debug_mode():
        app.sleep(2)
        widget.clicked.connect(lambda: logger.info("x"))
        Cursor.set_pos("screen", duration=1000)
        app.sleep(2)
        Cursor.set_pos(widget, how="right")



from __future__ import annotations

from collections.abc import Sequence
import contextlib
import logging

from typing import Any

from prettyqt import constants, core
from prettyqt.utils import classhelpers


logger = logging.getLogger(__name__)


class BaseDataclassModel(core.AbstractTableModel):
    DELEGATE_DEFAULT = "editor"

    def __init__(self, items: Sequence, **kwargs):
        super().__init__(**kwargs)
        self.items = items
        klasses = [type(i) for i in items]
        self.Class = classhelpers.lca_type(klasses)
        logger.debug(f"{type(self).__name__}: found common ancestor {self.Class}")
        self._fields = self.get_fields()
        self._field_names = list(self._fields.keys())

    def get_fields(self):
        return NotImplemented

    def columnCount(self, parent=None):
        return len(self._fields)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.VERTICAL, constants.DISPLAY_ROLE, _:
                instance = self.items[section]
                return type(instance).__name__
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self._field_names[section]

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        field_name = self._field_names[index.column()]
        instance = self.items[index.row()]
        value = getattr(instance, field_name)
        match role:
            case constants.DISPLAY_ROLE if not isinstance(value, bool):
                return repr(value)
            case constants.CHECKSTATE_ROLE if isinstance(value, bool):
                return self.to_checkstate(value)
            case constants.USER_ROLE | constants.EDIT_ROLE:
                return value

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        field_name = self._field_names[index.column()]
        instance = self.items[index.row()]
        match role:
            case constants.EDIT_ROLE | constants.USER_ROLE:
                with self.reset_model():
                    setattr(instance, field_name, value)
                return True
            case constants.CHECKSTATE_ROLE:
                with self.reset_model():
                    setattr(instance, field_name, bool(value))
                return True
        return False

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Override for AbstractitemModel base method."""
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return 0 if parent.isValid() else len(self.items)

    def flags(self, parent: core.ModelIndex) -> constants.ItemFlag:
        """Override.

        BaseClass implementation just tries to set attribute with same value to test
        if field is writable. If possible, subclasses should find a more efficient way.
        """
        if not parent.isValid():
            return super().flags(parent)
        field_name = self._field_names[parent.column()]
        instance = self.items[parent.row()]
        # need to cover not parent.isValid()?
        val = getattr(instance, field_name)
        with contextlib.suppress(Exception):
            setattr(instance, field_name, val)
            if isinstance(val, bool):
                return super().flags(parent) | constants.IS_CHECKABLE
            else:
                return super().flags(parent) | constants.IS_EDITABLE
        return super().flags(parent)



from __future__ import annotations

from prettyqt import location


class PlaceProposedSearchResult(
    location.PlaceSearchResultMixin, location.QPlaceProposedSearchResult
):
    """Represents a search result containing a proposed search."""



from __future__ import annotations

from prettyqt import bluetooth


class BluetoothHostInfo(bluetooth.QBluetoothHostInfo):
    """Encapsulates the details of a local QBluetooth device."""

    def set_address(self, address: bluetooth.QBluetoothAddress | int | str):
        """Set address to given value."""
        if isinstance(address, int | str):
            address = bluetooth.QBluetoothAddress(address)
        self.setAddress(address)

    def get_address(self) -> bluetooth.BluetoothAddress:
        """Get address."""
        return bluetooth.BluetoothAddress(self.address())


if __name__ == "__main__":
    address = BluetoothHostInfo()



from __future__ import annotations

import logging

from typing import Literal

from prettyqt import constants, core
from prettyqt.qt import QtGui
from prettyqt.utils import colors, datatypes


HighlightModeStr = Literal["column", "row", "both", "single"]

logger = logging.getLogger(__name__)


class HighlightMouseProxyModel(core.IdentityProxyModel):
    """Proxy model which highlights all cells with same row / column as mouse position.

    The proxy can work in four different modes.

    * column: The column the mouse is currently hovering over is highlighted.
    * row: The row the mouse is currently hovering over is highlighted.
    * both: Combination of column and row mode.
    * single: only the hovered cell is highlighted.
    """

    ID = "highlight_mouse"
    ICON = "mdi.cursor-default-click-outline"

    def __init__(
        self,
        parent: widgets.QAbstractItemView,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        mode: HighlightModeStr = "both",
        highlight_color: datatypes.ColorType = "red",
        **kwargs,
    ):
        self._mode = mode
        self._current_value = ...  # Sentinel value
        self._data_role = role
        self._current_column = None
        self._current_row = None
        self._highlight_color = colors.get_color(highlight_color).as_qt()
        super().__init__(parent, **kwargs)
        # TODO: this should be done in proxifier I think,
        # ItemModel shouldnt know about widget.
        parent.setMouseTracking(True)
        parent.entered.connect(self.cell_entered)
        parent.installEventFilter(self)

    def eventFilter(self, source, event):
        match event.type():
            case core.Event.Type.Leave:
                self._current_row = None
                self._current_column = None
                self.force_layoutchange()
        return False

    def cell_entered(self, index):
        self._current_row = index.row()
        self._current_column = index.column()
        self.force_layoutchange()

    def set_highlight_color(self, color: datatypes.ColorType):
        """Set color used for highlighting cells."""
        self._highlight_color = colors.get_color(color).as_qt()

    def get_highlight_color(self) -> QtGui.QColor:
        """Get color used for higlighting cells."""
        return self._highlight_color

    def set_highlight_mode(self, mode: HighlightModeStr):
        """Set highlight mode."""
        self._highlight_mode = mode

    def get_highlight_mode(self) -> HighlightModeStr:
        """Get highlight mode."""
        return self._highlight_mode

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role != constants.BACKGROUND_ROLE:
            return super().data(index, role)
        is_in_row = index.row() == self._current_row
        is_in_column = index.column() == self._current_column
        match self._mode:
            case "column" if is_in_column:
                return self._highlight_color
            case "row" if is_in_row:
                return self._highlight_color
            case "both" if is_in_row or is_in_column:
                return self._highlight_color
            case "single" if is_in_row and is_in_column:
                return self._highlight_color
            case _:
                return super().data(index, role)

    highlightMode = core.Property(
        str,
        get_highlight_mode,
        set_highlight_mode,
        doc="Highlight mode",
    )
    highlightColor = core.Property(
        QtGui.QColor,
        get_highlight_color,
        set_highlight_color,
        doc="Color to use for highlighting",
    )


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    table = debugging.example_table()
    table.proxifier.get_proxy("highlight_mouse")
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from collections.abc import Callable

from prettyqt import constants, core


class SelectionMixin:
    CHECKSTATE: dict[int, Callable] = {}  # column: identifier
    dataChanged: core.Signal

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.selected = {}

    def setData(
        self,
        index: core.ModelIndex,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if not index.isValid():
            return False
        elif role == constants.CHECKSTATE_ROLE:
            name = self._get_selection_id(index)
            self.selected[name] = not self.selected[name]
            self.dataChanged.emit(index, index)
            return True
        return super().setData(index, value, role)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return False
        if role == constants.CHECKSTATE_ROLE and index.column() == 0:
            name = self._get_selection_id(index)
            selected = self.selected.get(name, False)
            if name not in self.selected:
                self.selected[name] = selected
            return selected
        return super().data(index, role)

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        flags = super().flags(index)
        if index.column() in self.CHECKSTATE:
            return flags | constants.IS_CHECKABLE
        return flags

    def _get_selection_id(self, index: core.ModelIndex):
        item = index.data(constants.USER_ROLE)
        if id_fn := self.CHECKSTATE.get(index.column()):
            return id_fn(item)



from __future__ import annotations

from prettyqt import core, gui, widgets
from prettyqt.utils import get_repr


class KeySequenceEdit(widgets.WidgetMixin, widgets.QKeySequenceEdit):
    """Allows to input a QKeySequence."""

    value_changed = core.Signal(gui.QKeySequence)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keySequenceChanged.connect(self._on_value_change)
        self.setClearButtonEnabled(True)

    def __repr__(self):
        return get_repr(self, self.get_value())

    def _on_value_change(self, val):
        self.value_changed.emit(val)

    def set_value(self, value: str | gui.QKeySequence):
        if isinstance(value, str):
            value = gui.KeySequence.fromString(value)
        self.setKeySequence(value)

    def get_value(self) -> gui.QKeySequence:
        return self.keySequence()

    def is_valid(self) -> bool:
        return True

    def get_finishing_key_combinations(self) -> list[core.KeyCombination]:
        return [core.KeyCombination(i) for i in self.finishingKeyCombinations()]

    def get_key_combinations(self) -> list[core.KeyCombination]:
        return [core.KeyCombination(i) for i in self.keySequence()]


if __name__ == "__main__":
    app = widgets.app()
    widget = KeySequenceEdit()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.utils import datatypes


class PointEdit(widgets.Widget):
    """Simple Editor for QPoints."""

    value_changed = core.Signal(core.Point)

    def __init__(self, *args, object_name: str = "point_edit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_margin(0)
        self.spinbox_x = widgets.SpinBox(maximum=999999)
        self.spinbox_y = widgets.SpinBox(maximum=999999)
        layout = self.set_layout("horizontal", margin=0)
        layout.add(widgets.Label("x", alignment="center_right"))
        layout.add(self.spinbox_x)
        layout.add(widgets.Label("y", alignment="center_right"))
        layout.add(self.spinbox_y)
        self.spinbox_x.value_changed.connect(self._on_value_change)
        self.spinbox_y.value_changed.connect(self._on_value_change)

    def _on_value_change(self):
        self._value = self.get_value()
        self.value_changed.emit(self._value)

    def get_value(self) -> core.Point:
        return core.Point(
            self.spinbox_x.get_value(),
            self.spinbox_y.get_value(),
        )

    def set_value(self, value: datatypes.PointType):
        self._value = datatypes.to_point(value)
        self.spinbox_x.set_value(value.x())
        self.spinbox_y.set_value(value.y())

    value = core.Property(
        core.Point,
        get_value,
        set_value,
        user=True,
        doc="Current value",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = PointEdit(window_title="Test")
    widget.value_changed.connect(print)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import location
from prettyqt.utils import bidict


TYPE = bidict(
    unknown=location.QPlaceSearchResult.SearchResultType.UnknownSearchResult,
    place=location.QPlaceSearchResult.SearchResultType.PlaceResult,
    proposed_search=location.QPlaceSearchResult.SearchResultType.ProposedSearchResult,
)

TypeStr = Literal["unknown", "place", "proposed_search"]


class PlaceSearchResultMixin:
    def get_icon(self) -> location.PlaceIcon | None:
        icon = self.icon()
        return None if icon.isEmpty() else location.PlaceIcon(icon)

    def get_type(self) -> TypeStr:
        """Return result type.

        Returns:
            Result type
        """
        return TYPE.inverse[self.type()]


class PlaceSearchResult(PlaceSearchResultMixin, location.QPlaceSearchResult):
    """The base class for search results."""



"""A declarative framework for building dynamic apps with custom user interfaces."""

from __future__ import annotations

from prettyqt.qt.QtQuick import *  # noqa: F403

from .quicktextdocument import QuickTextDocument
from .quickrendercontrol import QuickRenderControl
from .quickimageresponse import QuickImageResponse
from .quickimageprovider import QuickImageProvider, QuickImageProviderMixin
from .quickasyncimageprovider import QuickAsyncImageProvider
from .quickitemgrabresult import QuickItemGrabResult
from .quickitem import QuickItem, QuickItemMixin
from .quickpainteditem import QuickPaintedItem
from .quickwindow import QuickWindow
from .quickview import QuickView
from .sgnode import SGNode
from prettyqt.qt import QtQuick

QT_MODULE = QtQuick

__all__ = [
    "SGNode",
    "QuickView",
    "QuickItem",
    "QuickItemMixin",
    "QuickWindow",
    "QuickPaintedItem",
    "QuickTextDocument",
    "QuickRenderControl",
    "QuickImageResponse",
    "QuickImageProvider",
    "QuickImageProviderMixin",
    "QuickAsyncImageProvider",
    "QuickItemGrabResult",
]



from __future__ import annotations

from prettyqt import widgets


class ItemDelegate(widgets.AbstractItemDelegateMixin, widgets.QItemDelegate):
    """Display and editing facilities for data items from a model."""



from __future__ import annotations

from typing import Literal

from typing_extensions import Self

from prettyqt import location
from prettyqt.utils import bidict, get_repr


OPERATION_TYPES = bidict(
    save_place=location.QPlaceIdReply.OperationType.SavePlace,
    remove_place=location.QPlaceIdReply.OperationType.RemovePlace,
    save_category=location.QPlaceIdReply.OperationType.SaveCategory,
    remove_category=location.QPlaceIdReply.OperationType.RemoveCategory,
)

OperationTypeStr = Literal[
    "save_place", "remove_place", "save_category", "remove_category"
]


class PlaceIdReply(location.PlaceReplyMixin, location.QPlaceIdReply):
    """Manages saving and removal operations of places and categories."""

    def __repr__(self):
        return get_repr(self, self.operationType())

    @classmethod
    def clone_from(cls, obj: location.QPlaceIdReply) -> Self:
        reply = cls(obj.operationType(), obj.parent())
        reply.setId(obj.id())
        return reply

    def get_operation_type(self) -> OperationTypeStr:
        """Get current operation_type.

        Returns:
            current operation_type
        """
        return OPERATION_TYPES.inverse[self.operationType()]


if __name__ == "__main__":
    reply = PlaceIdReply(location.QPlaceIdReply.OperationType.SavePlace)



"""Classes to make network programming easier and more portable."""

from __future__ import annotations

from prettyqt.qt.QtNetwork import *  # noqa: F403

from .networkcookie import NetworkCookie
from .networkcookiejar import NetworkCookieJar
from .networkreply import NetworkReply
from .networkrequest import NetworkRequest
from .networkproxy import NetworkProxy
from .hostaddress import HostAddress
from .networkaddressentry import NetworkAddressEntry
from .networkinterface import NetworkInterface
from .networkdatagram import NetworkDatagram
from .abstractsocket import AbstractSocket, AbstractSocketMixin
from .localsocket import LocalSocket
from .localserver import LocalServer
from .tcpserver import TcpServer
from .tcpsocket import TcpSocket
from .udpsocket import UdpSocket
from .httppart import HttpPart
from .httpmultipart import HttpMultiPart
from .networkaccessmanager import NetworkAccessManager
from prettyqt.qt import QtNetwork

QT_MODULE = QtNetwork

__all__ = [
    "NetworkProxy",
    "NetworkReply",
    "AbstractSocket",
    "AbstractSocketMixin",
    "LocalSocket",
    "HostAddress",
    "LocalServer",
    "TcpServer",
    "NetworkDatagram",
    "NetworkAddressEntry",
    "NetworkInterface",
    "TcpSocket",
    "UdpSocket",
    "HttpPart",
    "HttpMultiPart",
    "NetworkCookie",
    "NetworkCookieJar",
    "NetworkRequest",
    "NetworkAccessManager",
]



from __future__ import annotations

from prettyqt import constants, core, widgets


class SplitterHandle(widgets.WidgetMixin, widgets.QSplitterHandle):
    """Handle functionality for the splitter."""

    double_clicked = core.Signal(object)

    def __init__(
        self,
        orientation: constants.OrientationStr | constants.Orientation,
        parent: widgets.QSplitter,
        **kwargs,
    ):
        ori = constants.ORIENTATION.get_enum_value(orientation)
        super().__init__(ori, parent, **kwargs)

    @classmethod
    def setup_example(cls):
        w = widgets.Splitter("horizontal")
        return cls("horizontal", w)

    def mouseDoubleClickEvent(self, ev):
        self.double_clicked.emit(self)

    def set_orientation(
        self, orientation: constants.OrientationStr | constants.Orientation
    ):
        """Set the orientation of the slider.

        Args:
            orientation: orientation for the slider
        """
        self.setOrientation(constants.ORIENTATION.get_enum_value(orientation))

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Splitter()
    handle = SplitterHandle("horizontal", w)
    handle.show()
    app.exec()



from __future__ import annotations

import os

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import datatypes, get_repr


FormatStr = Literal["plaintext", "HTML", "markdown", "ODF"]


class TextDocumentWriter(QtGui.QTextDocumentWriter):
    def __repr__(self):
        return get_repr(self, self.device(), self.format())

    def get_format(self) -> FormatStr:
        return self.format().data().decode()  # type: ignore

    def set_format(self, fmt: FormatStr | bytes | core.QByteArray):
        new = fmt.encode() if isinstance(fmt, str) else fmt
        self.setFormat(new)

    def set_file_name(self, name: datatypes.PathType):
        path = name if isinstance(name, str) else os.fspath(name)
        self.setFileName(path)

    @classmethod
    def get_supported_document_formats(cls) -> list[str]:
        return [i.data().decode() for i in cls.supportedDocumentFormats()]

    @classmethod
    def serialize_document(
        cls,
        document: QtGui.QTextDocument,
        fmt: FormatStr | bytes | core.QByteArray = "ODF",
    ) -> bytes:
        buffer = core.Buffer()
        writer = cls()
        writer.setDevice(buffer)
        writer.set_format(fmt)  # ODF Format
        writer.write(document)
        return buffer.data().data()


if __name__ == "__main__":
    doc = TextDocumentWriter("a.text", b"plaintext")
    doc.set_format("UTF-8")
